package thkoeln.archilab.ecommerce.domainprimitives;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.Arrays;
import java.util.Locale;

import javax.persistence.Embeddable;

import lombok.EqualsAndHashCode;
import lombok.NoArgsConstructor;
import thkoeln.archilab.ecommerce.ShopException;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.MoneyType;

@Embeddable
@EqualsAndHashCode
@NoArgsConstructor
public class Money implements MoneyType {

    private Float amount;
    private String currency;

    private static final String[] ALLOWED_CURRENCIES = {"EUR", "CHF"};

    public Money(Float amount, String currency) {
        validateMoney(amount, currency);
        this.amount = amount;
        this.currency = currency;
    }

    public static void validateMoney(Float amount, String currency) throws ShopException {
        if (amount == null || amount < 0) {
            throw new ShopException("Amount must be non-null and non-negative");
        }
        if (currency == null || !(Arrays.asList(ALLOWED_CURRENCIES).contains(currency))) {
            throw new ShopException("Currency must be one of the allowed values: EUR, CHF");
        }
    }

    public static MoneyType of(Float amount, String currency) throws ShopException {
        validateMoney(amount, currency);
        return new Money(amount, currency);
    }

    public String toMoneyString() {
        if (amount == null || currency == null) {
            return "0.00 €";
        }
 // Use DecimalFormat to format the amount with comma as the decimal separator
        DecimalFormatSymbols symbols = new DecimalFormatSymbols(Locale.GERMANY);
        symbols.setDecimalSeparator('.');
        symbols.setGroupingSeparator('.'); // Set grouping separator in case we use it

        // Custom pattern without thousands separators
        DecimalFormat decimalFormat = new DecimalFormat("0.00", symbols);

        String formattedAmount = decimalFormat.format(amount);

        if(currency.equals("EUR")){
            return formattedAmount + " €";
        }
        else{
            return formattedAmount + " CHF";
        }
        
    }

    @Override
    public Float getAmount() {
        return amount;
    }

    @Override
    public String getCurrency() {
        return currency;
    }

    @Override
    public MoneyType add(MoneyType otherMoney) throws ShopException {
        validateOtherMoney(otherMoney);
        return new Money(this.amount + otherMoney.getAmount(), this.currency);
    }

    @Override
    public MoneyType subtract(MoneyType otherMoney) throws ShopException {
        validateOtherMoney(otherMoney);
        if (this.amount < otherMoney.getAmount()) {
            throw new ShopException("Cannot subtract more money than available");
        }
        return new Money(this.amount - otherMoney.getAmount(), this.currency);
    }

    @Override
    public MoneyType multiplyBy(int factor) throws ShopException {
        if (factor < 0) {
            throw new ShopException("Factor must be non-negative");
        }
        return new Money(this.amount * factor, this.currency);
    }

    @Override
    public boolean largerThan(MoneyType otherMoney) throws ShopException {
        validateOtherMoney(otherMoney);
        return this.amount > otherMoney.getAmount();
    }

    private void validateOtherMoney(MoneyType otherMoney) throws ShopException {
        if (otherMoney == null) {
            throw new ShopException("Other money must be non-null");
        }
        if (!this.currency.equals(otherMoney.getCurrency())) {
            throw new ShopException("Currencies must match");
        }
    }
}
