package thkoeln.archilab.ecommerce.solution.product.application;


import thkoeln.archilab.ecommerce.solution.product.domain.entities.Product;
import thkoeln.archilab.ecommerce.solution.product.domain.repositories.ProductRepository;
import thkoeln.archilab.ecommerce.usecases.ProductCatalogUseCases;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.MoneyType;
import thkoeln.archilab.ecommerce.ShopException;
import java.util.ArrayList;
import java.util.Optional;
import java.util.UUID;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;


@Service
public class ProductService implements ProductCatalogUseCases {

    private final ProductRepository productRepo;
    private final OrderProductChecker orderProductChecker;
    private final DepotProductChecker depotProductChecker;
    private final ShoppingCartProductChecker shoppingCartProductChecker;

    // 
    @Autowired
    public ProductService(ProductRepository productRepo, ShoppingCartProductChecker shoppingCartProductChecker
    , OrderProductChecker orderProductChecker, DepotProductChecker depotProductChecker){

        this.productRepo = productRepo;
        this.orderProductChecker = orderProductChecker;
        this.depotProductChecker = depotProductChecker;
        this.shoppingCartProductChecker = shoppingCartProductChecker;
    }

  

    @Override
    public void addProductToCatalog(UUID productId, String name, String description, Float size, MoneyType purchasePrice, MoneyType sellingPrice) {
        
/*|| purchasePrice <= 0 
 * sellingPrice <= 0 ||
 * 
 * 
 */

            if(productId == null || name == null || name.isEmpty() || description == null || description.isEmpty()
            || (size != null && size <= 0 ) || purchasePrice == null ||sellingPrice == null || purchasePrice.largerThan(sellingPrice)){
                throw new ShopException("addProductToCatalog() Input Data Validation failed");
            }

            Optional<Product> product = productRepo.findById(productId);
            if(!product.isEmpty()){
                throw new ShopException("Product already exists");
            }

            
            Product newProduct = new Product(productId, name, description, size, purchasePrice, sellingPrice);
            productRepo.save(newProduct);
    }

    @Override
    public void removeProductFromCatalog(UUID productId) {
        if(productId == null){
            throw new ShopException("removeProductFromCatalog() Input Data Validation failed");
        }

        Optional<Product> product = productRepo.findById(productId);
        if(product.isEmpty() || depotProductChecker.countProductInShoppingCarts(productId) > 0 ||  shoppingCartProductChecker.countProductInShoppingCarts(productId) > 0 ){
            throw new ShopException("removeProductFromCatalog Product does not exist or is still in Stock and can not be removed");
        }

        //OLD
        /*ArrayList<Order> orders = (ArrayList<Order>) orderRepo.findAll();

        for (Order order : orders) {
            List<OrderPart> orderParts = order.getOrderParts();
            
            for (OrderPart orderPart : orderParts) {    //maybe they want explicitly reserved or delivered but i cannot immagine why it would be allowed to delete paid and purchased
                if(orderPart.getProduct().getProductId() == productId){   
                    throw new ShopException("removeProductFromCatalog Product can not be removed - reserved or delivered etc");
                }
            }
        }

        productRepo.deleteById(productId);*/


 
        if (shoppingCartProductChecker.countProductInShoppingCarts(productId) > 0 || orderProductChecker.isProductInOrder(productId) ) {
            throw new ShopException("removeProductFromCatalog Product is in ShoppingCart or Orders and therefore cannot be removed"); 
        } else {
            productRepo.deleteById(productId);
        }
        
        
    }

    @Override
    public MoneyType getSellingPrice(UUID productId) {
        if(productId == null){
            throw new ShopException("getSellingPrice() Input Data Validation failed");
        }

        Optional<Product> product = productRepo.findById(productId);
        if(product.isEmpty() /*|| depotProductChecker.countProductInShoppingCarts(productId) > 0*/ ){
            throw new ShopException("getSellingPrice Product does not exist ");
        }

        return product.get().getSellingPrice();
    }

    @Override
    public void deleteProductCatalog() {
        ArrayList<Product> products = (ArrayList<Product>) productRepo.findAll();

        if(products != null && !products.isEmpty()) {
            productRepo.deleteAll(); 
        }
        
    }


}