package thkoeln.archilab.ecommerce.solution.shipmentunit.application;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Collectors;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import thkoeln.archilab.ecommerce.ShopException;
import thkoeln.archilab.ecommerce.solution.depot.domain.entities.Depot;
import thkoeln.archilab.ecommerce.solution.depot.domain.entities.StockLevel;
import thkoeln.archilab.ecommerce.solution.depot.domain.repositories.DepotRepository;
import thkoeln.archilab.ecommerce.solution.order.domain.Order;
import thkoeln.archilab.ecommerce.solution.order.domain.OrderRepository;
import thkoeln.archilab.ecommerce.solution.order.domain.valueObjects.orderPart.domain.OrderPart;
import thkoeln.archilab.ecommerce.usecases.ShipmentUnitUseCases;


@Service
public class ShipmentUnitService implements ShipmentUnitUseCases{

    private final OrderRepository orderRepo;
    private final DepotRepository depotRepo;
//    private final ShipmentUnitRepository unitRepo;

    @Autowired
    public ShipmentUnitService(OrderRepository orderRepo, DepotRepository depotRepo){
        //this.shipmentUnitRepo = shipmentUnitRepo;
        this.orderRepo = orderRepo;
        this.depotRepo = depotRepo;
//        this.unitRepo = unitRepo;
    }


    @Override
    public List<UUID> getContributingDepotsForOrder(UUID orderId) {
        if (orderId == null) {
            throw new ShopException("Order ID cannot be null.");
        }
    
        Optional<Order> orderTmp = orderRepo.findById(orderId);
        if (orderTmp.isEmpty()) {
            throw new ShopException("Order with ID " + orderId + " does not exist.");
        }
    
        Order order = orderTmp.get();
        List<OrderPart> orderParts = new ArrayList<>(order.getOrderParts());
        List<UUID> contributingDepots = new ArrayList<>();
    
        List<Depot> sortedDepots = getSortedDepotsByDistance(order);
    
        if (fulfillOrderFromSingleDepot(orderParts, contributingDepots, sortedDepots)) {
            return contributingDepots;
        }
    
        fulfillOrderFromMultipleDepots(order, orderParts, contributingDepots, sortedDepots);
    
        if (!orderParts.isEmpty()) {
            throw new ShopException("Not all parts of the order could be fulfilled.");
        }
    
        return contributingDepots;
    }
    
    private List<Depot> getSortedDepotsByDistance(Order order) {
        return depotRepo.findAll().stream()
                .sorted(Comparator.comparing(depot -> depot.getAddress().getPostalCode().distance(order.getUser().getAddress().getPostalCode())))
                .collect(Collectors.toList());
    }
    
    private boolean fulfillOrderFromSingleDepot(List<OrderPart> orderParts, List<UUID> contributingDepots, List<Depot> sortedDepots) {
        for (Depot depot : sortedDepots) {
            if (depotHasEnoughCapacity(depot, orderParts)) {
                contributingDepots.add(depot.getDepotId());
                updateStockLevels(depot, orderParts);
                return true;
            }
        }
        return false;
    }
    
    private void fulfillOrderFromMultipleDepots(Order order, List<OrderPart> orderParts, List<UUID> contributingDepots, List<Depot> sortedDepots) {
        int iterationOrderParts = 0;
    
        while (!orderParts.isEmpty() && iterationOrderParts < 100) {
            iterationOrderParts++;
    
            sortedDepots = getSortedDepotsBySupplyAndDistance(orderParts, sortedDepots, order);
    
            if (sortedDepots.isEmpty()) {
                break;
            }
    
            Depot selectedDepot = sortedDepots.get(0);
            boolean depotUsed = fulfillOrderPartially(selectedDepot, orderParts);
    
            if (depotUsed) {
                contributingDepots.add(selectedDepot.getDepotId());
            }
        }
    }
    
    private List<Depot> getSortedDepotsBySupplyAndDistance(List<OrderPart> orderParts, List<Depot> sortedDepots, Order order) {
        return sortedDepots.stream()
                .sorted(Comparator.comparing((Depot depot) -> {
                    int suppliedPartsCount = (int) orderParts.stream()
                            .filter(orderPart -> depot.findProductInStock(orderPart.getProduct().getProductId()) != null)
                            .count();
                    return -suppliedPartsCount; // Sort by most parts supplied, descending
                }).thenComparing(depot -> depot.getAddress().getPostalCode().distance(order.getUser().getAddress().getPostalCode())))
                .collect(Collectors.toList());
    }
    
    private boolean fulfillOrderPartially(Depot selectedDepot, List<OrderPart> orderParts) {
        boolean depotUsed = false;
    
        List<OrderPart> suppliedParts = new ArrayList<>();
        for (OrderPart orderPart : orderParts) {
            StockLevel stockLevel = selectedDepot.findProductInStock(orderPart.getProduct().getProductId());
            if (stockLevel != null && stockLevel.getStock() > 0) {
                int quantityToSupply = Math.min(orderPart.getQuantity(), stockLevel.getStock());
                if (quantityToSupply > 0) {
                    suppliedParts.add(new OrderPart( orderPart.getProduct(), quantityToSupply));
                    stockLevel.setStock(stockLevel.getStock() - quantityToSupply); // Update stock level
                    depotUsed = true;
    
                    // Update the original order part quantity
                    orderPart.setQuantity(orderPart.getQuantity() - quantityToSupply);
                }
            }
        }
    
        // Remove fully supplied parts from the order parts list
        orderParts.removeIf(orderPart -> orderPart.getQuantity() == 0);
    
        return depotUsed;
    }
    
    private void updateStockLevels(Depot depot, List<OrderPart> orderParts) {
        for (OrderPart orderPart : orderParts) {
            StockLevel stockLevel = depot.findProductInStock(orderPart.getProduct().getProductId());
            if (stockLevel != null && stockLevel.getStock() >= orderPart.getQuantity()) {
                stockLevel.setStock(stockLevel.getStock() - orderPart.getQuantity()); // Update stock level
            }
        }
    }
    
    private boolean depotHasEnoughCapacity(Depot depot, List<OrderPart> orderParts) {
        for (OrderPart orderPart : orderParts) {
            StockLevel stockLevel = depot.findProductInStock(orderPart.getProduct().getProductId());
            if (stockLevel == null || stockLevel.getStock() < orderPart.getQuantity()) {
                return false;
            }
        }
        return true;
    }


    


    @Override
    public Map<UUID, Integer> getShipmentUnitForOrderAndDepot(UUID orderId, UUID depotId) {
        if (orderId == null) {
            throw new ShopException("Order ID cannot be null.");
        }
        if (depotId == null) {
            throw new ShopException("Depot ID cannot be null.");
        }
    
        Optional<Order> orderTmp = orderRepo.findById(orderId);
    
        if (orderTmp.isEmpty()) {
            throw new ShopException("Order with ID " + orderId + " does not exist.");
        }
    
        Order order = orderTmp.get();
    
        Optional<Depot> requestedDepotTmp = depotRepo.findById(depotId);
        if (requestedDepotTmp.isEmpty()) {
            throw new ShopException("Depot with ID " + depotId + " does not exist.");
        }
    
        Depot requestedDepot = requestedDepotTmp.get();
    
        List<UUID> contributingDepots = getContributingDepotsForOrder(orderId);
        if (!contributingDepots.contains(depotId)) {
            return Collections.emptyMap();
        }
    
        Map<UUID, Integer> shipmentContents = new HashMap<>();
        List<OrderPart> orderParts = new ArrayList<>(order.getOrderParts());
    
        // Sort depots by proximity to the customer
        List<Depot> sortedDepots = contributingDepots.stream()
                .map(id -> depotRepo.findById(id).orElse(null))
                .filter(Objects::nonNull)
                .sorted(Comparator.comparing(depot -> depot.getAddress().getPostalCode().distance(order.getUser().getPersonalAddress().getPostalCode())))
                .collect(Collectors.toList());
    
        for (Depot depot : sortedDepots) {
            if (depot.getDepotId().equals(depotId)) {
                for (OrderPart orderPart : orderParts) {
                    StockLevel stockLevel = depot.findProductInStock(orderPart.getProduct().getProductId());
                    if (stockLevel != null) {
                        int stockToShip = Math.min(stockLevel.getStock(), orderPart.getQuantity());
                        if (stockToShip > 0) {
                            shipmentContents.put(orderPart.getProduct().getProductId(), stockToShip);
                        }
                    }
                }
                break;
            } else {
                for (OrderPart orderPart : orderParts) {
                    StockLevel stockLevel = depot.findProductInStock(orderPart.getProduct().getProductId());
                    if (stockLevel != null && stockLevel.getStock() > 0) {
                        int stockToShip = Math.min(stockLevel.getStock(), orderPart.getQuantity());
                        orderPart.setQuantity(orderPart.getQuantity() - stockToShip);
                    }
                }
            }
        }
    
        // Remove entries with zero values
        shipmentContents.entrySet().removeIf(entry -> entry.getValue() == 0);
    
        return shipmentContents;
    }
    








/* 
//product IDs as keys, and the amount of that products contained in 
    @Override
    public Map<UUID, Integer> getShipmentUnitForOrderAndDepot(UUID orderId, UUID depotId) {
        if (orderId == null) {
            throw new ShopException("Order ID cannot be null.");
        }
        if (depotId == null) {
            throw new ShopException("Depot ID cannot be null.");
        }
    
        Optional<Order> orderTmp = orderRepo.findById(orderId);
    
        if (orderTmp.isEmpty()) {
            throw new ShopException("Order with ID " + orderId + " does not exist.");
        }
    
        Order order = orderTmp.get();
    
        Optional<Depot> requestedDepotTmp = depotRepo.findById(depotId);
        if (requestedDepotTmp.isEmpty()) {
            throw new ShopException("Depot with ID " + depotId + " does not exist.");
        }
    
        Depot requestedDepot = requestedDepotTmp.get();
    
        List<UUID> contributingDepots = getContributingDepotsForOrder(orderId);
        if (!contributingDepots.contains(depotId)) {
            return Collections.emptyMap();
        }
    
        Map<UUID, Integer> shipmentContents = new HashMap<>();
        List<OrderPart> orderParts = new ArrayList<>(order.getOrderParts());
    
        // Sort depots by proximity to the customer
        List<Depot> sortedDepots = contributingDepots.stream()
                .map(id -> depotRepo.findById(id).orElse(null))
                .filter(Objects::nonNull)
                .sorted(Comparator.comparing(depot -> depot.getAddress().getPostalCode().distance(order.getUser().getPersonalAddress().getPostalCode())))
                .collect(Collectors.toList());

    
        for (Depot depot : sortedDepots) {
            if (depot.getDepotId().equals(depotId)) {
                for (OrderPart orderPart : orderParts) {
                    StockLevel stockLevel = depot.findProductInStock(orderPart.getProduct().getProductId());
                    if (stockLevel != null) {
                        int stockToShip = Math.min(stockLevel.getStock(), orderPart.getQuantity());
                        if (stockToShip > 0) {
                            shipmentContents.put(orderPart.getProduct().getProductId(), stockToShip);
                            orderParts.add(orderPart);

                        }
                    }
                }
                break;
            } else {
                for (OrderPart orderPart : orderParts) {
                    StockLevel stockLevel = depot.findProductInStock(orderPart.getProduct().getProductId());
                    if (stockLevel != null && stockLevel.getStock() > 0) {
                        int stockToShip = Math.min(stockLevel.getStock(), orderPart.getQuantity());
                        orderPart.setQuantity(orderPart.getQuantity() - stockToShip);
                        orderParts.add(orderPart);
                    }
                }
            }
//
//            ShipmentUnit unitNew = new ShipmentUnit();
//            unitNew.setDepot(depot);
//            unitNew.setOrder(order);
//            unitNew.setShipmentUnitId(UUID.randomUUID());
//
//            unitRepo.save(unitNew);
        }
    
        // Remove entries with zero values
        shipmentContents.entrySet().removeIf(entry -> entry.getValue() == 0);
    
       


        return shipmentContents;
    }
    */
    

    
    
    @Override
    public void deleteAllShipmentUnits() {
//        unitRepo.deleteAll(); //actually i had this but my shimentUnit Order and OrderPart made a hugeeeeeeeee mess
//        deleteAllDepots() dependency inversion
        depotRepo.deleteAll();
    }







}
