package thkoeln.archilab.ecommerce.solution.shoppingcart.application;

import java.util.List;
import java.util.UUID;

import org.json.JSONException;
import org.json.JSONObject;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseStatus;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.method.annotation.MethodArgumentTypeMismatchException;
import org.springframework.web.server.ResponseStatusException;

import thkoeln.archilab.ecommerce.domainprimitives.EmailAddress;
import thkoeln.archilab.ecommerce.solution.order.domain.valueObjects.orderPart.domain.OrderPart;
import thkoeln.archilab.ecommerce.solution.product.application.ProductNotFoundException;
import thkoeln.archilab.ecommerce.solution.shoppingcart.application.exceptions.ShoppingCartMethodNotAllowedException;
import thkoeln.archilab.ecommerce.solution.shoppingcart.application.exceptions.ShoppingCartNotFoundException;
import thkoeln.archilab.ecommerce.solution.shoppingcart.domain.entities.ShoppingCart;
import thkoeln.archilab.ecommerce.solution.user.application.UserDto;
import thkoeln.archilab.ecommerce.solution.user.application.UserDtoMapperHelper;
import thkoeln.archilab.ecommerce.solution.user.application.UserService;
import thkoeln.archilab.ecommerce.solution.user.application.exceptions.UserMethodNotAllowedException;
import thkoeln.archilab.ecommerce.solution.user.application.exceptions.UserNotFoundException;
import thkoeln.archilab.ecommerce.solution.user.domain.entities.User;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.EmailAddressType;



@RestController
@Transactional
public class ShoppingCartController {
    

     private final ShoppingCartService shoppingCartService;
     private final UserService userService;
     private final ShoppingCartDtoMapperHelper shoppingCartDtoMapperHelper;

    @Autowired
    public ShoppingCartController(ShoppingCartService shoppingCartService, UserService userService,ShoppingCartDtoMapperHelper shoppingCartDtoMapperHelper){
        this.shoppingCartService = shoppingCartService;
        this.userService = userService;
        this.shoppingCartDtoMapperHelper = shoppingCartDtoMapperHelper;

    }



  @PostMapping("/shoppingCarts/{shoppingCartId}/positions")
    public ResponseEntity<?> addProductToCart(  @PathVariable UUID shoppingCartId, @RequestBody String body ) {  //CartPositionDto positionDto
        try {
            JSONObject json;
            
                json = new JSONObject(body);
            

            CartPositionDto cartPositionDto = new CartPositionDto();
            cartPositionDto.setProductId(UUID.fromString(json.getString("productId")));
            cartPositionDto.setQuantity(json.getInt("quantity"));

            ShoppingCart cart;
            try {
                cart = shoppingCartService.getShoppingCartByCartId(shoppingCartId);

                if(cart == null)return ResponseEntity.status(HttpStatus.NOT_FOUND).build();
            } catch (Exception e) {
                if (e.getMessage()=="addProductToShoppingCart() Product does not exist or insufficient stock Exception") {
                    return ResponseEntity.status(HttpStatus.CONFLICT).body(e.getMessage());

                }
                return ResponseEntity.status(HttpStatus.NOT_FOUND).body(e.getMessage());
            }



            try {
                shoppingCartService.addProductToShoppingCart(EmailAddress.of( cart.getUser().getEmailAddress().toString() ), cartPositionDto.getProductId(), cartPositionDto.getQuantity());
                //return ResponseEntity.ok(cartPositionDto);

                return ResponseEntity.status(HttpStatus.CREATED).body(cartPositionDto);//.build();

            } catch (ShoppingCartNotFoundException e) {
                if (e.getMessage()=="addProductToShoppingCart() Product does not exist or insufficient stock Exception") {
                    return ResponseEntity.status(HttpStatus.CONFLICT).body(e.getMessage());

                }
                return ResponseEntity.status(HttpStatus.NOT_FOUND).body(e.getMessage());
            } catch (ProductNotFoundException e) {
                if (e.getMessage()=="addProductToShoppingCart() Product does not exist or insufficient stock Exception") {
                    return ResponseEntity.status(HttpStatus.CONFLICT).body(e.getMessage());

                }
                return ResponseEntity.status(HttpStatus.NOT_FOUND).body(e.getMessage());
            } catch (Exception e) {
                if(e.getMessage()== "addProductToShoppingCart() Product insufficient stock Exception"){
                    return ResponseEntity.status(HttpStatus.CONFLICT).body(e.getMessage());

                 }
                else if(e.getMessage()=="addProductToShoppingCart() Product not found"){
                    return ResponseEntity.status(HttpStatus.NOT_FOUND).body(e.getMessage());
                }
                return ResponseEntity.status(HttpStatus.UNPROCESSABLE_ENTITY).body(e.getMessage());
            }

        } catch (JSONException e) {
            return ResponseEntity.status(HttpStatus.INTERNAL_SERVER_ERROR).body(e.getMessage());
        }
    }






     @DeleteMapping("/shoppingCarts/{shoppingCartId}/positions/{productId}")
    public ResponseEntity<Void> deleteProductFromShoppingCart(
            @PathVariable UUID shoppingCartId,
            @PathVariable UUID productId) throws ShoppingCartMethodNotAllowedException {
        

        ShoppingCart shoppingCart = null;

        try {
            shoppingCart = shoppingCartService.getShoppingCartByCartId(shoppingCartId);
        } catch (Exception e) {
            throw new ResponseStatusException(HttpStatus.CONFLICT, e.getMessage());
        }


        if(shoppingCartId == null || productId == null) return ResponseEntity.status(HttpStatus.METHOD_NOT_ALLOWED).build();

        




        shoppingCart = shoppingCartService.getShoppingCartByCartId(shoppingCartId);
        if(shoppingCart.getCartItems() == null){
            return ResponseEntity.status(HttpStatus.NOT_FOUND).build();
        } 
        


        int amountToFree = 0;
        for (OrderPart part : shoppingCart.getCartItems()) {
            if(part.getProduct().getProductId().equals(productId)){
                amountToFree = part.getQuantity();
                break;
            }
        }
        
        //if(amountToFree == 0){ return ResponseEntity.status(HttpStatus.NOT_FOUND).build();}


        try {
            shoppingCartService.removeProductFromShoppingCart(EmailAddress.of(shoppingCart.getUser().getEmailAddress().toString()), productId, amountToFree);
            
            
            //return ResponseEntity.ok();
            return ResponseEntity.status(HttpStatus.OK).build();



        } catch (ShoppingCartNotFoundException e) {
            throw new ResponseStatusException(HttpStatus.NOT_FOUND, e.getMessage());
        } catch (Exception e) {

            if(e.getMessage()== "removeProductFromShoppingCart() quantity < 1"){
                throw new ResponseStatusException(HttpStatus.CONFLICT, e.getMessage());
             }
             if(e.getMessage()== "removeProductFromShoppingCart() Product does not exist Exception"){
                throw new ResponseStatusException(HttpStatus.NOT_FOUND, e.getMessage());
             }
            throw new ResponseStatusException(HttpStatus.CONFLICT, e.getMessage());
        }
    }



    @PostMapping("/shoppingCarts/{shoppingCartId}/checkout")
    public ResponseEntity<ShoppingCartDto> checkoutShoppingCart(@PathVariable UUID shoppingCartId)
            throws ShoppingCartMethodNotAllowedException {
        
        ShoppingCart shoppingCart;
        try {
            shoppingCart = shoppingCartService.getShoppingCartByCartId(shoppingCartId);
        } catch (Exception e) {
            throw new ShoppingCartNotFoundException(e.getMessage());
        }


        try {
            ShoppingCartDto cartDto = shoppingCartDtoMapperHelper.mapToDto(shoppingCartService.getShoppingCartByCartId(shoppingCartId)); 
            
            UUID orderId = shoppingCartService.checkout(shoppingCart.getUser().getEmailAddress());
            cartDto.setId(orderId);

            
            return ResponseEntity.status(HttpStatus.CREATED).body(cartDto); // Return the UUID of the checked-out shopping cart
        } catch (ShoppingCartNotFoundException e) {
            throw new ResponseStatusException(HttpStatus.NOT_FOUND, e.getMessage());
        } catch (Exception e) {
            if(e.getMessage() == "checkout() Empty Shopping Cart Exception")
            {
                throw new ResponseStatusException(HttpStatus.CONFLICT, e.getMessage());
            }
            throw new ResponseStatusException(HttpStatus.NOT_FOUND, e.getMessage());
        }
    }






    //Working below
   /*  @GetMapping("/shoppingCarts/{cartId}")
    public ResponseEntity<ShoppingCartDto> getOneShoppingCartByCartId(@PathVariable UUID cartId) 
            throws ShoppingCartMethodNotAllowedException {
        ShoppingCart shoppingCart;
        try {
            shoppingCart = shoppingCartService.getShoppingCartByCartId(cartId);
        } catch (Exception e) {
            throw new ShoppingCartNotFoundException(e.getMessage());
        }

        ShoppingCartDto shoppingCartDto = ShoppingCartDtoMapperHelper.mapToDto(shoppingCart);
        return ResponseEntity.ok(shoppingCartDto);
    }*/




        
    @GetMapping("/shoppingCarts")
    public ResponseEntity<ShoppingCartDto> getOneShoppingCartByUserId(@RequestParam(required = false) String userId) 
            throws ShoppingCartMethodNotAllowedException {


                if(userId == null || userId == ""){throw new ShoppingCartMethodNotAllowedException("No User with userId = " + userId);}

        /*  User user;
        try {
            user = (User)userService.getUserData(UUID.fromString(userId));
        } catch (Exception e) {
            throw new UserNotFoundException("No User with userId = " + userId);
        }

        if (user == null) {
            throw new UserNotFoundException("No User with userId = " + userId);
        }*/

        ShoppingCart shoppingCart;
        try {
            shoppingCart = shoppingCartService.getShoppingCartByUserId(UUID.fromString(userId));
        } catch (Exception e) {
            throw new ShoppingCartNotFoundException(e.getMessage());
        }

        ShoppingCartDto shoppingCartDto = ShoppingCartDtoMapperHelper.mapToDto(shoppingCart);
        return ResponseEntity.ok(shoppingCartDto);
    }


}
