package thkoeln.archilab.ecommerce.solution.shoppingcart.application;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import thkoeln.archilab.ecommerce.solution.product.domain.repositories.ProductRepository;
import thkoeln.archilab.ecommerce.solution.shipmentunit.domain.ShipmentUnit;
import thkoeln.archilab.ecommerce.solution.shoppingcart.domain.entities.ShoppingCart;
import thkoeln.archilab.ecommerce.solution.shoppingcart.domain.repositories.ShoppingCartRepository;
import thkoeln.archilab.ecommerce.solution.user.application.UserShoppingCartHelper;
import thkoeln.archilab.ecommerce.solution.user.domain.entities.User;
import thkoeln.archilab.ecommerce.solution.user.domain.entities.UserShoppingCartCreator;
import thkoeln.archilab.ecommerce.solution.user.domain.repositories.UserRepository;
import thkoeln.archilab.ecommerce.ShopException;
import thkoeln.archilab.ecommerce.domainprimitives.EmailAddress;
import thkoeln.archilab.ecommerce.usecases.ShoppingCartUseCases;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.EmailAddressType;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.MoneyType;
import thkoeln.archilab.ecommerce.solution.order.domain.Order;
import thkoeln.archilab.ecommerce.solution.order.domain.OrderRepository;
import thkoeln.archilab.ecommerce.solution.order.domain.valueObjects.orderPart.domain.OrderPart;
import thkoeln.archilab.ecommerce.solution.product.application.ShoppingCartProductChecker;
import thkoeln.archilab.ecommerce.solution.product.domain.entities.Product;

@Service
public class ShoppingCartService implements ShoppingCartUseCases, ShoppingCartProductChecker, UserShoppingCartHelper, UserShoppingCartCreator{    
    

    private final OrderRepository orderRepo;
    private final UserRepository userRepo;
    private final ProductRepository productRepo;
    private final ShoppingCartRepository cartRepo;
   // private final DepotProductChecker depotChecker;

    //private DepotService depotService;

    //, DepotService depotService, , DepotProductChecker depotChecker
    @Autowired
    public ShoppingCartService(OrderRepository orderRepo, UserRepository userRepo, ProductRepository productRepo, ShoppingCartRepository cartRepo){
        this.orderRepo = orderRepo;
        this.userRepo = userRepo;
        this.productRepo = productRepo; 
        this.cartRepo = cartRepo; 
        //this.depotService = depotService; 
        //this.depotChecker = depotChecker; 
    }

    @Override
    public void createShoppingCartForUser(User user) {
        ShoppingCart cart = new ShoppingCart();
        cart.setId(UUID.randomUUID());
        cart.setUser(user);
        cart.setCartItems(new ArrayList<OrderPart>());
        cartRepo.save(cart);
    }

 
    public ShoppingCart getShoppingCartByCartId(UUID cartId){
        Optional<ShoppingCart> cart = cartRepo.findById(cartId);
        if(cart == null || cart.isEmpty() ) throw new ShopException("no Cart found by CartId = " + cartId);

        return cart.get();
    }

    public ShoppingCart getShoppingCartByUserId(UUID userid){


        Optional<User> user = userRepo.findById(userid);
        if(user == null || user.isEmpty() ) throw new ShopException("no user found by userid = " + userid);


        ShoppingCart cart = cartRepo.findByUser(user.get());
        if(cart == null){
            //OrderPart newOrderPart = new OrderPart(product, quantity);
            ShoppingCart newOrder = new ShoppingCart(user.get());
            cart = cartRepo.save(newOrder);
        }

        return cart;
    }



    //Old stuff below

    //ToDo refactor check if user exist to method

    @Override
    public boolean isEmpty(EmailAddressType userEmailAddress) {
        if(userEmailAddress == null ) throw new ShopException("isEmpty() Input Exception"); //|| userEmailAddress.isEmpty() 

        User user = userRepo.findByEmailAddress((EmailAddress)EmailAddress.of(userEmailAddress.toString()));
        if (user == null) throw new ShopException("isEmpty() User does not exist Exception");

        ShoppingCart cart = cartRepo.findByUser(user);
        
        return cart == null ? true : cart.getCartItems().isEmpty();
    }

    @Override
    public void emptyAllShoppingCarts() {
        List<ShoppingCart> carts = cartRepo.findAll();

        for (ShoppingCart shoppingCart : carts) {
            List<OrderPart> newCart = new ArrayList<OrderPart>();
            shoppingCart.setCartItems(newCart);
            cartRepo.save(shoppingCart);
        }
    }



    @Override
    public void addProductToShoppingCart(EmailAddressType userEmailAddress, UUID productId, int quantity) {
        if (userEmailAddress == null || productId == null || quantity < 0) throw new ShopException("addProductToShoppingCart() Input Exception");

        User user = userRepo.findByEmailAddress((EmailAddress)EmailAddress.of(userEmailAddress.toString()));
        if (user == null) throw new ShopException("addProductToShoppingCart() User does not exist Exception");

        Product product = productRepo.findByProductId(productId);

        if(product == null)throw new ShopException("addProductToShoppingCart() Product not found");
        
        int totalStock = product.getTotalStock();
        int reservedStock = getReservedStockInShoppingCarts(productId);


        if (product == null ||  totalStock - reservedStock < quantity) throw new ShopException("addProductToShoppingCart() Product insufficient stock Exception");

        ShoppingCart cart = cartRepo.findByUser(user);
        if (cart == null) {
            // Neuer Einkaufswagen erstellen und OrderPart hinzufügen
            OrderPart orderPart = new OrderPart( product, quantity);
            ShoppingCart newCart = new ShoppingCart(user, orderPart);
            cartRepo.save(newCart);
            return;
        }

        List<OrderPart> cartItems = cart.getCartItems();
        boolean sameProductFound = false;
        for (OrderPart cartItem : cartItems) {
            if (cartItem.getProduct().getProductId().equals(product.getProductId())) {
                cartItem.setQuantity(cartItem.getQuantity() + quantity);
                sameProductFound = true;
                break;
            }
        }

        if (!sameProductFound) {
            OrderPart newOrderPart = new OrderPart();
            newOrderPart.setProduct(product);
            newOrderPart.setQuantity(quantity);
            cart.getCartItems().add(newOrderPart);
        }

        cartRepo.save(cart);
    }

    @Override
    public void removeProductFromShoppingCart(EmailAddressType userEmailAddress, UUID productId, int quantity) {
        if(userEmailAddress == null || productId ==  null ) throw new ShopException("removeProductFromShoppingCart() Input Exception");
        
        Product product = productRepo.findByProductId(productId);
        if (product == null) throw new ShopException("removeProductFromShoppingCart() Product does not exist Exception");

        if( quantity < 1) throw new ShopException("removeProductFromShoppingCart() quantity < 1");

        User user = userRepo.findByEmailAddress((EmailAddress)EmailAddress.of(userEmailAddress.toString()));
        if (user == null) throw new ShopException("removeProductFromShoppingCart() User does not exist Exception");

       



        ShoppingCart cart = cartRepo.findByUser(user);  
        List<OrderPart> cartItems = cart.getCartItems();


        if(cartItems == null || cartItems.size()<1) throw new ShopException("removeProductFromShoppingCart() Empty Shopping Cart Exception");

        OrderPart orderPartToRemove = null;

        for (OrderPart orderPart : cartItems) {
            if (orderPart.getProduct().getProductId().equals(productId)) {
                orderPartToRemove = orderPart;
                break;
            }
        }
        
        if(orderPartToRemove == null) {
            throw new ShopException("Nothing to Remove");
        }

        // If the product is not in the cart or quantity to remove exceeds available quantity, throw exception
        if ( orderPartToRemove.getQuantity() < quantity) {
            throw new ShopException("Product with ID " + productId + " is not in the cart in the requested quantity");
        }

        // Reduce the quantity of the product in the order part
        int newTotalQuantity = orderPartToRemove.getQuantity() - quantity;

        // If quantity becomes zero, remove the order part
        if(newTotalQuantity == 0){
            cartItems.remove(orderPartToRemove);
        }else if(newTotalQuantity < 0){
            throw new ShopException("Product with ID " + productId + " is not in the cart in the requested quantity");
        }else{
            orderPartToRemove.setQuantity(newTotalQuantity);
        }

        cart.setCartItems(cartItems);


        cartRepo.save(cart);
    }

    @Override
    public Map<UUID, Integer> getShoppingCartAsMap(EmailAddressType userEmailAddress) {
        HashMap<UUID, Integer> output = new HashMap<UUID, Integer>();

        if(userEmailAddress == null) throw new ShopException("getShoppingCartAsMap() Input Exception");
        
        User user = userRepo.findByEmailAddress((EmailAddress)EmailAddress.of(userEmailAddress.toString()));
        if (user == null) throw new ShopException("getShoppingCartAsMap() User does not exist Exception"); 

        ShoppingCart cart = cartRepo.findByUser(user);  
        List<OrderPart> cartItems = cart.getCartItems();
        
        if(cartItems.size()<1) return output;

        for (OrderPart orderPart : cartItems) {
            output.put(orderPart.getProduct().getProductId(), orderPart.getQuantity());
        }

        return output;
    }

    @Override
    public MoneyType getShoppingCartAsMoneyValue(EmailAddressType userEmailAddress) {
        if(userEmailAddress == null) throw new ShopException("getShoppingCartAsMoneyValue() Input Exception");
        
        User user = userRepo.findByEmailAddress((EmailAddress)EmailAddress.of(userEmailAddress.toString()));
        if (user == null) throw new ShopException("getShoppingCartAsMoneyValue() User does not exist Exception");

       
        ShoppingCart cart = cartRepo.findByUser(user);  
        /*List<OrderPart> cartItems = cart.getCartItems();
       
        if(cartItems.size()<1) return 0;


        float output = 0;
        for (OrderPart orderPart : cartItems) {
            output += orderPart.getQuantity() * orderPart.getProduct().getSellingPrice();
        }*/

        return cart.getCartTotalPrice();
    }

    @Override
    public int getReservedStockInShoppingCarts(UUID productId) {

        if(productId ==  null) throw new ShopException("getReservedStockInShoppingCarts() Input Exception");
        
        Product product = productRepo.findByProductId(productId);
        if (product == null) throw new ShopException("getReservedStockInShoppingCarts() Product does not exist Exception");

        int totalStockReserved = 0;
        List<ShoppingCart> carts = cartRepo.findAll();

        for (ShoppingCart cart : carts) {
            List<OrderPart> orderParts = cart.getCartItems(); 
            for (OrderPart orderPart : orderParts) {
                if(orderPart.getProduct().getProductId().equals(productId) ){
                    totalStockReserved += orderPart.getQuantity();
                }
            }
        }
        return totalStockReserved;
    }

    

    


    @Override
    public UUID checkout(EmailAddressType userEmailAddress) {
        if(userEmailAddress == null ) throw new ShopException("checkout() Input Exception");
        
        User user = userRepo.findByEmailAddress((EmailAddress)EmailAddress.of(userEmailAddress.toString()));
        if (user == null) throw new ShopException("checkout() User does not exist Exception");

        ShoppingCart cart = cartRepo.findByUser(user);  
        if (cart == null) throw new ShopException("checkout() ShoppingCart does not exist Exception");
        List<OrderPart> cartItems = cart.getCartItems();
       
        if(cartItems.size()<1) throw new ShopException("checkout() Empty Shopping Cart Exception");

        //ToDo stock correction

        for (OrderPart cartItem : cartItems) {
            
            Product product = productRepo.findByProductId(cartItem.getProduct().getProductId());
            if(product !=null){
                product.setTotalStock(product.getTotalStock()-cartItem.getQuantity());
                productRepo.save(product);
            }
        }

        //depotRepo

        Order order = new Order(cart.getUser(), cart.getCartItems());
        order = orderRepo.save(order);

        
       
        cartItems = new ArrayList<OrderPart>();
        cart.setCartItems(cartItems);
        cartRepo.save(cart);


        



       // orders.get(0).setStatus(Order.OrderStatus.DELIVERED);
       return order.getOrderId();
    }








     @Override
    public int countProductInShoppingCarts(UUID productId) {
        return getReservedStockInShoppingCarts(productId);
    }

    @Override
    public void removeProductInShoppingCarts(UUID productId, int quantity) {
        if(productId == null ) throw new ShopException("reduceProductInShoppingCarts() Error"); 

        Product product = productRepo.findByProductId(productId);
        if(product == null) throw new ShopException("changeStockTo() Error, Product does not exist.");
                
        List<ShoppingCart> carts = cartRepo.findAll();

        int amountToFree = quantity;
        if(amountToFree > 0){
            for (ShoppingCart cart : carts) {
                List<OrderPart> orderParts = cart.getCartItems();
                
                for (OrderPart orderPart : orderParts) {
                    if(orderPart.getProduct().getProductId().equals(productId)){  

                        int orderPartQuantity = orderPart.getQuantity();
                        if(orderPartQuantity- amountToFree > 0){
                            orderPart.setQuantity(orderPartQuantity- amountToFree);
                            amountToFree = 0;
                            cartRepo.save(cart);
                            break;
                        }else{
                            orderPart.setQuantity(0);
                            amountToFree = amountToFree - orderPartQuantity;
                            cartRepo.save(cart);
                        }
                    }
                }
                
                if(amountToFree == 0){
                    break;
                } 
            }
        }
        //product.setStock(removedQuantity);
    }



    @Override
    public void deleteAllCarts() {
        cartRepo.deleteAll();
    }


    

  /*   @Override
    public int countProductInShoppingCarts(UUID productId) {
        
        return getReservedStockInShoppingCarts(productId);
    }

    @Override   //return amount of freed products
    public void removeProductInShoppingCarts(UUID productId, int quantity) {

        if(productId == null ) throw new ShopException("reduceProductInShoppingCarts() Error"); 

        Product product = productRepo.findByProductId(productId);
        if(product == null) throw new ShopException("changeStockTo() Error, Product does not exist.");
                
        List<ShoppingCart> carts = cartRepo.findAll();

        int amountToFree = quantity;
        if(amountToFree > 0){
            for (ShoppingCart cart : carts) {
                List<OrderPart> orderParts = cart.getCartItems();
                
                for (OrderPart orderPart : orderParts) {
                    if(orderPart.getProduct().getProductId() == productId){  

                        int orderPartQuantity = orderPart.getQuantity();
                        if(orderPartQuantity- amountToFree > 0){
                            orderPart.setQuantity(orderPartQuantity- amountToFree);
                            amountToFree = 0;
                            break;
                        }else{
                            orderPart.setQuantity(0);
                            amountToFree = amountToFree - orderPartQuantity;
                        }
                    }
                }
                
                if(amountToFree == 0) break;
            }
        }
        //product.setStock(removedQuantity);
        
    }

*/


    /* 
    @Override
    public boolean isProductInShoppingCart(UUID productId) {
        List<ShoppingCart> allCarts = cartRepo.findAll();
        
        for (ShoppingCart shoppingCart : allCarts) {
            List<OrderPart> cartItems = shoppingCart.getCartItems();
            for (OrderPart cartItem : cartItems) {
                if(cartItem.getProduct().getProductId() == productId){
                    return true;
                }
            }
        }

        return false;
    }*/


}
