package thkoeln.archilab.ecommerce.usecases;

import thkoeln.archilab.ecommerce.ShopException;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.PersonalAddressType;

import java.util.UUID;


/**
 * This interface contains methods needed in the context of managing depots, and handling
 * the shop stock, i.e. adding and removing products in the depot.
 */
public interface DepotUseCases {
    /**
     * Adds a new depot to the shop
     * @param address
     * @param name
     * @return the id of the new depot
     * @throws ShopException if ...
     *      - address is null
     *      - name is null or empty
     */
    public UUID addNewDepot( PersonalAddressType address, String name );


    /**
     * Deletes all depots from the shop. Intended for testing purposes.
     * @throws ShopException if the products catalog is not empty
     */
    public void deleteAllDepots();


    /**
     * Adds a certain quantity of a given product to the stock
     * @param depotId
     * @param productId
     * @param addedQuantity
     * @throws ShopException if ...
     *      - depotId is null
     *      - the depot with that id does not exist
     *      - productId is null
     *      - the product with that id does not exist
     *      - addedQuantity < 0
     */
    public void addToStock( UUID depotId, UUID productId, int addedQuantity );


    /**
     * Removes a certain quantity of a given product from the stock.
     * If the new total quantity is lower than the currently reserved products, some of currently reserved products
     * (in the users' shopping carts) are removed. This means that some of the reserved products are lost for
     * the user. (This is necessary because there probably was a mistake in the stock management, a mis-counting,
     * or some of the products were stolen from the depot, are broken, etc.)
     * @param depotId
     * @param productId
     * @param removedQuantity
     * @throws ShopException if ...
     *      - depotId is null
     *      - the depot with that id does not exist
     *      - productId is null
     *      - the product with that id does not exist
     *      - removedQuantity < 0
     *      - the removed quantity is greater than the current stock and the currently reserved products together
     */
    public void removeFromStock( UUID depotId, UUID productId, int removedQuantity );


    /**
     * Changes the total quantity of a given product in the stock.
     * If the new total quantity is lower than the currently reserved products, some of currently reserved products
     * (in the users' shopping carts) are removed. This means that some of the reserved products are lost for
     * the user. (This is necessary because there probably was a mistake in the stock management, a mis-counting,
     * or some of the products were stolen from the depot, are broken, etc.)
     * @param depotId
     * @param productId
     * @param newTotalQuantity
     * @throws ShopException if ...
     *      - depotId is null
     *      - the depot with that id does not exist
     *      - productId is null
     *      - the product with that id does not exist
     *      - newTotalQuantity < 0
     */
    public void changeStockTo( UUID depotId, UUID productId, int newTotalQuantity );


    /**
     * Get the current stock of a given product in one specific depot.
     * @param depotId
     * @param productId
     * @return the current total stock of the product
     * @throws ShopException if ...
     *      - depotId is null
     *      - the depot with that id does not exist
     *      - productId is null
     *      - the product with that id does not exist
     */
    public int getAvailableStock( UUID depotId, UUID productId );


    /**
     * Get the current total stock of a given product, across all depots, and including the currently
     * reserved products in shopping carts.
     * @param productId
     * @return the current total stock of the product
     * @throws ShopException if ...
     *      - productId is null
     *      - the product with that id does not exist
     */
    public int getAvailableStock( UUID productId );
}
