package thkoeln.archilab.ecommerce.usecases;

import thkoeln.archilab.ecommerce.ShopException;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.EmailAddressType;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.MoneyType;

import java.util.Map;
import java.util.UUID;

/**
 * This interface contains methods needed in the context of use cases handling the shopping cart of
 * a user.
 */
public interface ShoppingCartUseCases {
    /**
     * Adds a product to the shopping cart of a user
     *
     * @param userEmailAddress
     * @param productId
     * @param quantity
     * @throws ShopException if ...
     *      - the user with the given emailAddress does not exist,
     *      - the product does not exist,
     *      - the quantity <= 0,
     *      - the product is not available in the requested quantity
     */
    public void addProductToShoppingCart( EmailAddressType userEmailAddress, UUID productId, int quantity );


    /**
     * Removes a product from the shopping cart of a user
     *
     * @param userEmailAddress
     * @param productId
     * @param quantity
     * @throws ShopException if ...
     *      - userEmailAddress is null,
     *      - the user with the given emailAddress does not exist,
     *      - the product does not exist
     *      - the quantity <= 0,
     *      - the product is not in the shopping cart in the requested quantity
     */
    public void removeProductFromShoppingCart( EmailAddressType userEmailAddress, UUID productId, int quantity );


    /**
     * Returns a map showing which products are in the shopping cart of a user and how many of each product
     *
     * @param userEmailAddress
     * @return the shopping cart of the user (map is empty if the shopping cart is empty)
     * @throws ShopException if ...
     *      - userEmailAddress is null,
     *      - the user with the given emailAddress does not exist
     */
    public Map<UUID, Integer> getShoppingCartAsMap( EmailAddressType userEmailAddress );


    /**
     * Returns the current value of all products in the shopping cart of a user
     *
     * @param userEmailAddress
     * @return the value of shopping cart of the user
     * @throws ShopException if ...
     *      - userEmailAddress is null,
     *      - the user with the given emailAddress does not exist
     */
    public MoneyType getShoppingCartAsMoneyValue( EmailAddressType userEmailAddress );


    /**
     * Get the number units of a specific product that are currently reserved in the shopping carts of all users
     * @param productId
     * @return the number of reserved products of that type in all shopping carts
     * @throws ShopException
     *      - productId is null
     *      - if the product id does not exist
     */
    public int getReservedStockInShoppingCarts( UUID productId );


    /**
     * Checks if the shopping cart of a user is empty
     *
     * @param userEmailAddress
     * @return true if the shopping cart is empty, false otherwise
     * @throws ShopException if ...
     *    - userEmailAddress is null
     *    - the user with the given email address does not exist
     */
    public boolean isEmpty( EmailAddressType userEmailAddress );


    /**
     * Checks out the shopping cart of a user. This means that the products in the shopping cart
     * are removed from the stock. The shopping cart is emptied.
     *
     * @param userEmailAddress
     * @return the id of the order that was created
     * @throws ShopException if ...
     *      - userEmailAddress is null
     *      - the user with the given email address does not exist
     *      - the shopping cart is empty
     *      - the products in the shopping cart are not available in the requested quantity
     */
    public UUID checkout( EmailAddressType userEmailAddress );


    /**
     * Empties all shopping carts for all users
     */
    public void emptyAllShoppingCarts();
}
