package thkoeln.archilab.ecommerce.usecases.masterdata;

import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import thkoeln.archilab.ecommerce.usecases.ProductCatalogUseCases;
import thkoeln.archilab.ecommerce.usecases.DepotUseCases;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.PersonalAddressType;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.MoneyType;

import java.util.*;

/**
 * This is a test helper class that initializes and registers products (but without stock)
 * in the system, using the given interface(s).
 */
@Component
@Slf4j
@SuppressWarnings("PMD")
public class ProductAndStockMasterDataInitializer {

    private ProductCatalogUseCases productCatalogUseCases;
    private DepotUseCases depotUseCases;
    private static Random random = new Random();

    public static final String EUR = "EUR";

    @Autowired
    public ProductAndStockMasterDataInitializer( ProductCatalogUseCases productCatalogUseCases,
                                                                DepotUseCases depotUseCases ) {
        this.productCatalogUseCases = productCatalogUseCases;
        this.depotUseCases = depotUseCases;
    }

    // Contains initialization data for product instances, and their stock in the depot
    // The data is structured as follows:
    // { UUID, name, description, weight, inPrice, outPrice, depots }
    // where "depots" is a string with numbers between 0 and 9, representing the depots
    // where the product is stored.
    //
    // There is a total of PRODUCT_NUMOF products in the data. The index is (for simplicity)
    // noted as a number at the beginning of the product name, like "0-TCD-34 v2.1" or "1-EFG-56".
    //
    // With regard to where the stock is stored, the following rules apply:
    // - products 0, 1, 2, and 3 have fixed quantities of 0, 10, 20, and 30, respectively. For simplicity,
    //   they are ONLY available in depot 0.

    public static final int PRODUCT_NUMOF = 15;
    public static final Object[][] PRODUCT_DATA = new Object[][]{
            {UUID.randomUUID(), "0-TCD-34 v2.1", "Universelles Verbindungsstück für den einfachen Hausgebrauch bei der Schnellmontage",
                    1.5f, FactoryMethodInvoker.instantiateMoney( 5.0f, EUR ),
                    FactoryMethodInvoker.instantiateMoney( 10.0f, EUR ), "0"},
            {UUID.randomUUID(), "1-EFG-56", "Hochleistungsfähiger Kondensator für elektronische Schaltungen",
                    0.3f, FactoryMethodInvoker.instantiateMoney( 2.5f, EUR ),
                    FactoryMethodInvoker.instantiateMoney( 4.0f, EUR ), "0"},
            {UUID.randomUUID(), "2-MNP-89ff", "Langlebiger und robuster Motor für industrielle Anwendungen",
                    7.2f, FactoryMethodInvoker.instantiateMoney( 50.0f, EUR ),
                    FactoryMethodInvoker.instantiateMoney( 80.0f, EUR ), "0"},
            {UUID.randomUUID(), "3-Gh-25", "Kompakter und leichter Akku für mobile Geräte",
                    null, FactoryMethodInvoker.instantiateMoney( 6.0f, EUR ),
                    FactoryMethodInvoker.instantiateMoney( 8.0f, EUR ), "0"},
            {UUID.randomUUID(), "4-MultiBeast2", "Vielseitiger Adapter für verschiedene Steckertypen",
                    null, FactoryMethodInvoker.instantiateMoney( 0.6f, EUR ),
                    FactoryMethodInvoker.instantiateMoney( 1.0f, EUR ), "0"},
            {UUID.randomUUID(), "5-ABC-99 v4.2", "Leistungsstarker Prozessor für Computer und Server",
                    1.0f, FactoryMethodInvoker.instantiateMoney( 150.0f, EUR ),
                    FactoryMethodInvoker.instantiateMoney( 250.0f, EUR ), "0"},
            {UUID.randomUUID(), "6-Stuko22", "Ersatzteil Spitze für Präzisionswerkzeug zum Löten und Schrauben",
                    null, FactoryMethodInvoker.instantiateMoney( 0.3f, EUR ),
                    FactoryMethodInvoker.instantiateMoney( 0.5f, EUR ), "0"},
            {UUID.randomUUID(), "7-Btt2-Ah67", "Kraftstoffeffiziente und umweltfreundliche Hochleistungsbatterie",
                    6.0f, FactoryMethodInvoker.instantiateMoney( 80.0f, EUR ),
                    FactoryMethodInvoker.instantiateMoney( 120.0f, EUR ), "123"},
            {UUID.randomUUID(), "8-JKL-67", "Wasserdichtes Gehäuse",
                    3.0f, FactoryMethodInvoker.instantiateMoney( 1.0f, EUR ),
                    FactoryMethodInvoker.instantiateMoney( 1.2f, EUR ), "467"},
            {UUID.randomUUID(), "9-MNO-55-33", "Modulares Netzteil für flexible Stromversorgung",
                    5.5f, FactoryMethodInvoker.instantiateMoney( 25.0f, EUR ),
                    FactoryMethodInvoker.instantiateMoney( 45.0f, EUR ), "578"},
            {UUID.randomUUID(), "10-PQR-80", "Effizienter Kühler für verbesserte Wärmeableitung",
                    4.0f, FactoryMethodInvoker.instantiateMoney( 20.0f, EUR ),
                    FactoryMethodInvoker.instantiateMoney( 35.0f, EUR ), "567"},
            {UUID.randomUUID(), "11-STU-11 Ld", "Hochwertiger Grafikchip für leistungsstarke PCs",
                    null, FactoryMethodInvoker.instantiateMoney( 200.0f, EUR ),
                    FactoryMethodInvoker.instantiateMoney( 350.0f, EUR ), "478"},
            {UUID.randomUUID(), "12-VWX-90 FastWupps", "Schnellladegerät für eine Vielzahl von Geräten",
                    null, FactoryMethodInvoker.instantiateMoney( 15.0f, EUR ),
                    FactoryMethodInvoker.instantiateMoney( 25.0f, EUR ), "5"},
            {UUID.randomUUID(), "13-YZZ-22 v1.8", "Leichter und stabiler Rahmen aus Aluminium",
                    3.5f, FactoryMethodInvoker.instantiateMoney( 60.0f, EUR ),
                    FactoryMethodInvoker.instantiateMoney( 100.0f, EUR ), "78"},
            {UUID.randomUUID(), "14-Nosedive", "Klammer zum Verschließen der Nase beim Tauchen",
                    5.0f, FactoryMethodInvoker.instantiateMoney( 2.0f, EUR ),
                    FactoryMethodInvoker.instantiateMoney( 5.0f, EUR ), "457"}
    };


    /**
     * Used for creating invalid input data for the products in the tests.
     */
    public enum InvalidReason {
        NULL, EMPTY;

        public Object getInvalidValue( Object originalValue ) {
            switch (this) {
                case NULL:
                    return null;
                case EMPTY:
                    return "";
                default:
                    return null;
            }
        }
    }

    public void addAllProducts() {
        log.info( "Adding all products to the catalog." );
        for ( Object[] productData : PRODUCT_DATA ) {
            addProductDataToCatalog( productData );
        }
    }

    public void addProductDataToCatalog( Object[] productData ) {
        productCatalogUseCases.addProductToCatalog( (UUID) productData[0], (String) productData[1], (String) productData[2],
                (Float) productData[3], (MoneyType) productData[4], (MoneyType) productData[5] );
        log.info( "... added " + productData[1] + " with id " + productData[0] + " to the catalog." );
    }

    public Object[] getProductDataInvalidAtIndex( int index, InvalidReason reason ) {
        Object[] productData = PRODUCT_DATA[1];
        Object[] productDataInvalid = new Object[productData.length];
        System.arraycopy( productData, 0, productDataInvalid, 0, productData.length );
        productDataInvalid[index] = productData[index].getClass().cast(
                reason.getInvalidValue( productData[index] ) );
        return productDataInvalid;
    }


    // These personal addresss are used for the depots. The depot name will equal
    // the postal code of the personal address. Their index number will be visible in the house number.
    // The depots are used as such:
    // - depot 0 is holds all products 0 - 6, and is used for all tests where multiple
    //   shipment units are irrelevant.
    // - depots 1 - 3 are used for the proximity tests, where you can deliver products 7 to
    //   to a user from the closest depot.
    // - depots 4 - 8 are used for the tests where you need to deliver products 8 - 14 in
    //   the most cost-efficient way, as multiple shipment units.
    // - depot 9 is empty.
    public final static int DEPOT_NUMOF = 10;
    public final static PersonalAddressType[] DEPOT_ADDRESS = new PersonalAddressType[]{
            FactoryMethodInvoker.instantiatePersonalAddress(
                    "Stapelallee 0", "Potsdam",
                    FactoryMethodInvoker.instantiatePostalCode( "14476" ) ),
            FactoryMethodInvoker.instantiatePersonalAddress(
                    "Lagerhausstr. 1", "Viertelstadt",
                    FactoryMethodInvoker.instantiatePostalCode( "02345" ) ),
            FactoryMethodInvoker.instantiatePersonalAddress(
                    "Speicherplatz 2", "Viertelstadt",
                    FactoryMethodInvoker.instantiatePostalCode( "02313" ) ),
            FactoryMethodInvoker.instantiatePersonalAddress(
                    "Ablageweg 3", "Reichswürgen",
                    FactoryMethodInvoker.instantiatePostalCode( "44923" ) ),
            FactoryMethodInvoker.instantiatePersonalAddress(
                    "Paketstellenallee 4", "Düsseldorf",
                    FactoryMethodInvoker.instantiatePostalCode( "40588" ) ),
            FactoryMethodInvoker.instantiatePersonalAddress(
                    "Kaputte-Sachen-Straße 5", "Düren",
                    FactoryMethodInvoker.instantiatePostalCode( "52355" ) ),
            FactoryMethodInvoker.instantiatePersonalAddress(
                    "Aufbewahrungsweg 6", "Viernheim",
                    FactoryMethodInvoker.instantiatePostalCode( "68519" ) ),
            FactoryMethodInvoker.instantiatePersonalAddress(
                    "Paketallee 7", "Baden-Baden",
                    FactoryMethodInvoker.instantiatePostalCode( "76532" ) ),
            FactoryMethodInvoker.instantiatePersonalAddress(
                    "Sendenstr. 8", "Senden",
                    FactoryMethodInvoker.instantiatePostalCode( "89250" ) ),
            FactoryMethodInvoker.instantiatePersonalAddress(
                    "Schickweg 9", "Hohenroth",
                    FactoryMethodInvoker.instantiatePostalCode( "97618" ) )
    };
    public final static UUID[] DEPOT_ID = new UUID[DEPOT_NUMOF];

    public void addAllDepots() {
        log.info( "Adding all depots." );
        for ( int i = 0; i < DEPOT_NUMOF; i++ ) {
            DEPOT_ID[i] = depotUseCases.addNewDepot(
                    DEPOT_ADDRESS[i], DEPOT_ADDRESS[i].getPostalCode().toString() );
            log.info( "... added depot Nr. " + i + " at postal code " + DEPOT_ADDRESS[i].getPostalCode()
                    + " with id " + DEPOT_ID[i] );
        }
    }


    // These data structures contain the stock of the products in the depots.
    // PRODUCT_STOCK is a map product id -> Integer[DEPOT_NUMOF].
    // The Integer[DEPOT_NUMOF] contains the stock of the product in each of
    // the depots.
    //
    // The following rules apply:
    // - product 0 is out of stock
    // - product 1 / 2 / 3 have fixed quantities of 10 / 20 / 30 respectively, all ONLY in depot 0
    // - product 4 / 5 / 6 have a random stock between 30 and 130, also all ONLY in depot 0
    //   (these are the products used for tests on how to add and remove stock)
    // - the others have a random stock between 30 and 130, distributed over several
    //   depots. Here we follow this convention for simplicity:
    //   - Assume that the product is available in <n> depots. Then the first <n-1> depots
    //     in the list (in ascending sequence) contain 3, and all the remaining stock is in the
    //     last depot.

    public static final Map<UUID, Integer[]> PRODUCT_STOCK = new HashMap<>();

    static {
        // products 0, 1, 2, and 3 have fixed quantities of 0, 10, 20, and 30.
        PRODUCT_STOCK.put( (UUID) PRODUCT_DATA[0][0],
                getStockDistribution( 0, (String) PRODUCT_DATA[0][6] ) );
        PRODUCT_STOCK.put( (UUID) PRODUCT_DATA[1][0],
                getStockDistribution( 10, (String) PRODUCT_DATA[1][6] ) );
        PRODUCT_STOCK.put( (UUID) PRODUCT_DATA[2][0],
                getStockDistribution( 20, (String) PRODUCT_DATA[2][6] ) );
        PRODUCT_STOCK.put( (UUID) PRODUCT_DATA[3][0],
                getStockDistribution( 30, (String) PRODUCT_DATA[3][6] ) );

        // The other products have a random stock between 30 and 130,
        for ( int i = 4; i < PRODUCT_NUMOF; i++ ) {
            Integer totalNumber = random.nextInt( 100 ) + 30;
            Integer[] stockInDepots =
                    getStockDistribution( totalNumber, (String) PRODUCT_DATA[i][6] );
            PRODUCT_STOCK.put( (UUID) PRODUCT_DATA[i][0], stockInDepots );
        }
    }


    /**
     * This method creates a random stock distribution for the given product.
     *
     * @param totalQuantity - the total number of products in the depots
     * @param zeroToNine  - a string with numbers between 0 and 9, representing the depots
     * @return an Integer array with the stock distribution for the product, according to
     * the rules described above.
     */
    private static Integer[] getStockDistribution( Integer totalQuantity, String zeroToNine ) {
        Integer[] stockInDepots = new Integer[DEPOT_NUMOF];
        for ( int i = 0; i < DEPOT_NUMOF; i++ ) stockInDepots[i] = 0;
        TreeSet<Integer> depotIndices = getDepotIndices( zeroToNine );
        int numOfIndices = depotIndices.size();
        int currentIndexNumber = 0;
        int currentQuantity = totalQuantity;
        for ( Integer depotIndex : depotIndices ) {
            currentIndexNumber++;
            if ( currentIndexNumber < numOfIndices ) {
                stockInDepots[depotIndex] = 3;
                currentQuantity -= 3;
            } else {
                stockInDepots[depotIndex] = currentQuantity;
            }
        }
        return stockInDepots;
    }


    private static TreeSet<Integer> getDepotIndices( String zeroToNine ) {
        TreeSet<Integer> depotIndices = new TreeSet<>();
        for ( int i = 0; i < zeroToNine.length(); i++ ) {
            depotIndices.add( Integer.parseInt( zeroToNine.substring( i, i + 1 ) ) );
        }
        return depotIndices;
    }


    public void addAllStock() {
        log.info( "Adding all stocks to the depots." );
        for ( Object[] productData : PRODUCT_DATA ) {
            Integer[] stockInDepots =
                    PRODUCT_STOCK.get( productData[0] );
            for ( int iDepot = 0; iDepot < DEPOT_NUMOF; iDepot++ ) {
                if ( stockInDepots[iDepot] > 0 )
                    depotUseCases.addToStock(
                            DEPOT_ID[iDepot], (UUID) productData[0],
                            stockInDepots[iDepot] );
            }
        }
    }


    public Integer findDepotIndex( UUID depotId ) {
        for ( int i = 0; i < DEPOT_NUMOF; i++ ) {
            if ( DEPOT_ID[i].equals( depotId ) ) {
                return i;
            }
        }
        return null;
    }

    /**
     * This method is used to get the total price of a shopping cart, given as a map of product id -> quantity.
     * @param quantityProductMap
     * @return the total price of the shopping cart, as Float
     */
    public Float getTotalSellingPrice( Map<UUID, Integer> quantityProductMap ) {
        Float totalPrice = 0.0f;
        for ( UUID productId : quantityProductMap.keySet() ) {
            Integer quantity = quantityProductMap.get( productId );
            int productIndex = -1;
            for ( int i = 0; i < PRODUCT_NUMOF; i++ ) {
                if ( PRODUCT_DATA[i][0].equals( productId ) ) {
                    productIndex = i;
                    break;
                }
            }
            totalPrice += ((MoneyType) PRODUCT_DATA[productIndex][5]).getAmount() * quantity;
        }
        return totalPrice;
    }
}
