package thkoeln.archilab.ecommerce.test;

import com.fasterxml.jackson.databind.ObjectMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.stereotype.Component;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.MvcResult;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.ResultMatcher;
import thkoeln.archilab.ecommerce.test.restdtos.IdDTO;
import thkoeln.archilab.ecommerce.test.restdtos.QuantityDTO;
import thkoeln.archilab.ecommerce.usecases.ProductCatalogUseCases;
import thkoeln.archilab.ecommerce.usecases.DepotUseCases;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.EmailAddressType;
import thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer;

import java.util.Map;
import java.util.UUID;

import static org.springframework.http.MediaType.APPLICATION_JSON;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.jsonPath;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;


public class ShoppingCartRESTHelper {
    private MockMvc mockMvc;
    private ProductCatalogUseCases productCatalogUseCases;
    private DepotUseCases depotUseCases;

    private static final ResultMatcher CREATED = status().isCreated();
    private static final ResultMatcher OK = status().isOk();
    private ProductAndStockMasterDataInitializer productAndStockMasterDataInitializer;

    @Autowired
    public ShoppingCartRESTHelper( MockMvc mockMvc,
                                    ProductCatalogUseCases productCatalogUseCases,
                                    DepotUseCases depotUseCases ) {
        this.mockMvc = mockMvc;
        this.productCatalogUseCases = productCatalogUseCases;
        this.depotUseCases = depotUseCases;
        productAndStockMasterDataInitializer = new ProductAndStockMasterDataInitializer(
                productCatalogUseCases, depotUseCases );
    }

    /**
     * Executes a GET query on a shopping cart and returns its id, with the user's email address as a
     * starting point.
     * @param emailAddress
     * @return
     * @throws Exception
     */
    public UUID getQueryShoppingCart( EmailAddressType emailAddress ) throws Exception {
        return getQueryShoppingCart( emailAddress, null );
    }


    /**
     * Executes a GET query on a shopping cart and returns its id, with the user's email address as a
     * starting point.
     * @param emailAddress
     * @param quantityMap A map of products and their quantities in the shopping cart. (If null, then don't check.)
     * @return The id of the user's shopping cart.
     * @throws Exception
     */
    public UUID getQueryShoppingCart( EmailAddressType emailAddress, Map<UUID, Integer> quantityMap )
            throws Exception {
        // first query the users API to get the proper userId
        UUID userId;
        String userUri = "/users?emailAddress=" + emailAddress.toString();
        MvcResult userGetResult = mockMvc.perform( get( userUri ) )
                .andExpect( status().isOk() )
                .andReturn();
        ObjectMapper objectMapper = new ObjectMapper();
        IdDTO idDTO = objectMapper.readValue( userGetResult.getResponse().getContentAsString(), IdDTO.class );
        userId = idDTO.getId();

        // then query the shopping cart API and extract the shopping cart id
        String shoppingCartUri = "/shoppingCarts?userId=" + userId.toString();
        ResultActions resultActions = mockMvc.perform( get( shoppingCartUri ) )
                .andExpect( status().isOk() );
        idDTO = objectMapper.readValue( resultActions.andReturn().getResponse().getContentAsString(), IdDTO.class );
        UUID shoppingCartId = idDTO.getId();

        // additional checks, if desired
        if ( quantityMap != null ) {
            Float totalSellingPrice = productAndStockMasterDataInitializer
                    .getTotalSellingPrice( quantityMap );
            String totalSellingPriceString = String.format( "%.2f €", totalSellingPrice );
            resultActions
                    .andExpect( jsonPath( "$.totalSellingPrice" ).value( totalSellingPriceString ) );
            checkQuantityMap( quantityMap, resultActions );
        }
        return shoppingCartId;
    }


    public void checkQuantityMap( Map<UUID, Integer> quantityMap, ResultActions resultActions ) throws Exception {
        int numOfProducts = quantityMap.size();
        resultActions.andExpect( jsonPath( "$.shoppingCartPositions.length()" ).value( numOfProducts ) );
        for ( Map.Entry<UUID, Integer> entry : quantityMap.entrySet() ) {
            UUID productId = entry.getKey();
            Integer quantity = entry.getValue();
            resultActions.andExpect( jsonPath( "$.shoppingCartPositions.[?(@.productId == '" + productId + "')].quantity" )
                    .value( quantity ) );
        }
    }


    /**
     * Adds a product to a shopping cart.
     * @param shoppingCartId Id of the shopping cart.
     * @param productId Id of the product.
     * @param quantity Quantity of the product.
     * @throws Exception
     */
    public void addProductToShoppingCart( UUID shoppingCartId, UUID productId, Integer quantity )
            throws Exception {
        addProductToShoppingCart( shoppingCartId, productId, quantity, null );
    }


    /**
     * Adds a product to a shopping cart.
     * @param shoppingCartId Id of the shopping cart.
     * @param productId Id of the product.
     * @param quantity Quantity of the product.
     * @param expectedStatus expected status of the operation (null = CREATED)
     * @throws Exception
     */
    public void addProductToShoppingCart( UUID shoppingCartId, UUID productId, Integer quantity,
                                                  ResultMatcher expectedStatus ) throws Exception {
        ResultMatcher status = expectedStatus == null ? CREATED : expectedStatus;
        ObjectMapper objectMapper = new ObjectMapper();
        QuantityDTO dto = new QuantityDTO( productId, quantity );
        String quantityJson = objectMapper.writeValueAsString( dto );
        mockMvc.perform( post( "/shoppingCarts/" + shoppingCartId + "/positions" )
                        .contentType( APPLICATION_JSON ).content( quantityJson ) )
                .andExpect( status );
    }


    /**
     * Deletes a product from a shopping cart.
     * @param shoppingCartId Id of the shopping cart.
     * @param productId Id of the product.
     * @throws Exception
     */
    public void deleteProductFromShoppingCart( UUID shoppingCartId, UUID productId )
            throws Exception {
        deleteProductFromShoppingCart( shoppingCartId, productId, null );
    }


    /**
     * Deletes a product from a shopping cart.
     * @param shoppingCartId Id of the shopping cart.
     * @param productId Id of the product.
     * @param expectedStatus expected status of the operation (null = OK)
     * @throws Exception
     */
    public void deleteProductFromShoppingCart( UUID shoppingCartId, UUID productId,
                                                      ResultMatcher expectedStatus ) throws Exception {
        ResultMatcher status = expectedStatus == null ? OK : expectedStatus;
        String uri = "/shoppingCarts/" + shoppingCartId + "/positions/" + productId;
        mockMvc.perform( delete( uri ) ).andExpect( status );
    }


    /**
     * Checks out a shopping cart, and returns the id of the created order.
     * @param shoppingCartId
     * @return
     * @throws Exception
     */
    public UUID checkout( UUID shoppingCartId ) throws Exception {
        return checkout( shoppingCartId, null );
    }


    /**
     * Checks out a shopping cart, and returns the id of the created order.
     * @param shoppingCartId
     * @param expectedStatus expected status of the operation (null = CREATED)
     * @return
     * @throws Exception
     */
    public UUID checkout( UUID shoppingCartId, ResultMatcher expectedStatus ) throws Exception {
        ResultMatcher status = expectedStatus == null ? CREATED : expectedStatus;
        String uri = "/shoppingCarts/" + shoppingCartId + "/checkout";
        MvcResult result = mockMvc.perform( post( uri ) ).andExpect( status ).andReturn();
        if ( !status.equals( CREATED ) ) return null;
        ObjectMapper objectMapper = new ObjectMapper();
        IdDTO idDTO = objectMapper.readValue( result.getResponse().getContentAsString(), IdDTO.class );
        return idDTO.getId();
    }
}
