package thkoeln.archilab.ecommerce.test.regression;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer;
import thkoeln.archilab.ecommerce.usecases.masterdata.UserMasterDataInitializer;
import thkoeln.archilab.ecommerce.usecases.masterdata.Purgatory;
import thkoeln.archilab.ecommerce.usecases.*;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.EmailAddressType;

import java.util.List;
import java.util.Map;
import java.util.UUID;

import static org.junit.Assert.fail;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer.PRODUCT_DATA;
import static thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer.DEPOT_ID;
import static thkoeln.archilab.ecommerce.usecases.masterdata.UserMasterDataInitializer.USER_EMAIL;

/**
 * This test checks the management of stock across depots. It focuses on the
 * cases where only one depot can serve the shopping cart of a user.
 */
@SpringBootTest
public class OneShipmentTest {
    @Autowired
    private UserRegistrationUseCases userRegistrationUseCases;
    @Autowired
    private ShoppingCartUseCases shoppingCartUseCases;
    @Autowired
    private ProductCatalogUseCases productCatalogUseCases;
    @Autowired
    private DepotUseCases depotUseCases;
    @Autowired
    private ShipmentUnitUseCases shipmentUnitUseCases;
    @Autowired
    private Purgatory purgatory;

    private UserMasterDataInitializer userMasterDataInitializer;
    private ProductAndStockMasterDataInitializer productAndStockMasterDataInitializer;

    private Integer[][] closestDepotIndex = new Integer[][]{
            {2},          // user 0 at 02314 => shipment from 02313 (2)
            {1, 2},       // user 1 at 02368 => shipment from 02345 (1) or 02313 (2)
            {1, 2},       // user 2 at 12345 => shipment from 02345 (1) or 02313 (2)
            {3},          // user 3 at 31463 => shipment from 44923 (3)
            {1, 2, 3},    // user 4 at 72162 => shipment from 44923 (3) or 02345 (1) or 02313 (2)
            {1, 2}        // user 5 at 82195 => shipment from from 02345 (1) or 02313 (2)
    };

    @BeforeEach
    public void setUp() {
        purgatory.deleteEverything();

        userMasterDataInitializer = new UserMasterDataInitializer( userRegistrationUseCases );
        userMasterDataInitializer.registerAllUsers();

        productAndStockMasterDataInitializer = new ProductAndStockMasterDataInitializer(
                productCatalogUseCases, depotUseCases );
        productAndStockMasterDataInitializer.addAllProducts();
        productAndStockMasterDataInitializer.addAllDepots();
        productAndStockMasterDataInitializer.addAllStock();
    }


    /**
     * Different users always buy the same product (no. 7). The product is available in
     * depot no. {1, 2, 3}. The test checks if the correct (closest) depot is
     * selected for each user.
     */
    @Test
    public void testClosestShipmentUnit() {
        // given
        UUID productId7 = (UUID) PRODUCT_DATA[7][0];
        UUID[] orderIds = new UUID[6];

        // when
        for ( int iUser = 0; iUser <= 5; iUser++ ) {
            EmailAddressType userEmail = USER_EMAIL[iUser];
            shoppingCartUseCases.addProductToShoppingCart( userEmail, productId7, 1 );
            orderIds[iUser] = shoppingCartUseCases.checkout( userEmail );
        }

        // then
        for ( int iUser = 0; iUser <= 5; iUser++ ) {
            EmailAddressType userEmail = USER_EMAIL[iUser];
            List<UUID> depots = shipmentUnitUseCases
                    .getContributingDepotsForOrder( orderIds[iUser] );
            assertEquals( 1, depots.size() );
            assertDepotMatches( iUser, userEmail, depots );

            Map<UUID, Integer> productQuantity = shipmentUnitUseCases
                    .getShipmentUnitForOrderAndDepot( orderIds[iUser],
                            depots.get( 0 ) );
            assertEquals( 1, productQuantity.size() );
            assertEquals( productId7, productQuantity.keySet().iterator().next() );
            assertEquals( 1, productQuantity.values().iterator().next() );
        }
    }


    private void assertDepotMatches( int iUser, EmailAddressType userEmail,
                                                List<UUID> depots ) {
        Integer[] closestDepotIndices = closestDepotIndex[iUser];
        for ( Integer closestDepotIndex : closestDepotIndices ) {
            if ( depots.contains( DEPOT_ID[closestDepotIndex] ) ) {
                return;
            }
        }
        String msg = "Wrong depot selected for user " + userEmail + ".\n" +
                "Expected one of the following depot indices: {";
        for ( int i = 0; i < closestDepotIndices.length; i++ ) {
            msg += closestDepotIndices[i];
            if ( i < closestDepotIndices.length - 1 ) msg += ", ";
        }
        msg += "}, but got the following indices: {";
        for ( int i = 0; i < depots.size(); i++ ) {
            UUID id = depots.get( i );
            Integer idIndex = productAndStockMasterDataInitializer.findDepotIndex( id );
            msg += idIndex;
            if ( i < depots.size() - 1 ) msg += ", ";
        }
        msg += "}";
        fail( msg );
    }

}
