package thkoeln.archilab.ecommerce.test.regression;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import thkoeln.archilab.ecommerce.ShopException;
import thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer;
import thkoeln.archilab.ecommerce.usecases.masterdata.UserMasterDataInitializer;
import thkoeln.archilab.ecommerce.usecases.masterdata.Purgatory;
import thkoeln.archilab.ecommerce.usecases.ShoppingCartUseCases;
import thkoeln.archilab.ecommerce.usecases.ProductCatalogUseCases;
import thkoeln.archilab.ecommerce.usecases.UserRegistrationUseCases;
import thkoeln.archilab.ecommerce.usecases.DepotUseCases;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.EmailAddressType;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.MoneyType;

import java.util.UUID;

import static org.junit.Assert.assertEquals;
import static org.junit.jupiter.api.Assertions.assertDoesNotThrow;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer.PRODUCT_DATA;
import static thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer.DEPOT_ID;
import static thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer.InvalidReason.EMPTY;
import static thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer.InvalidReason.NULL;
import static thkoeln.archilab.ecommerce.usecases.masterdata.UserMasterDataInitializer.USER_EMAIL;


@SpringBootTest
public class ProductCatalogTest {
    @Autowired
    private UserRegistrationUseCases userRegistrationUseCases;
    @Autowired
    private ShoppingCartUseCases shoppingCartUseCases;
    @Autowired
    private ProductCatalogUseCases productCatalogUseCases;
    @Autowired
    private DepotUseCases depotUseCases;
    @Autowired
    private Purgatory purgatory;

    private UserMasterDataInitializer userMasterDataInitializer;
    private ProductAndStockMasterDataInitializer productAndStockMasterDataInitializer;


    @BeforeEach
    public void setUp() {
        purgatory.deleteEverything();

        userMasterDataInitializer = new UserMasterDataInitializer( userRegistrationUseCases );
        userMasterDataInitializer.registerAllUsers();

        productAndStockMasterDataInitializer = new ProductAndStockMasterDataInitializer(
                productCatalogUseCases, depotUseCases );
        productAndStockMasterDataInitializer.addAllProducts();
        productAndStockMasterDataInitializer.addAllDepots();
    }


    @Test
    public void testAddProductToCatalog() {
        // given
        // when
        MoneyType sellingPrice = productCatalogUseCases.getSellingPrice( (UUID) PRODUCT_DATA[4][0] );

        // then
        assertEquals( PRODUCT_DATA[4][5], sellingPrice );
    }


    @Test
    public void testAddProductWithInvalidData() {
        // given
        Object[][] invalidProductData = {
            productAndStockMasterDataInitializer.getProductDataInvalidAtIndex( 0, NULL ),     // id
            productAndStockMasterDataInitializer.getProductDataInvalidAtIndex( 1, NULL ),     // name
            productAndStockMasterDataInitializer.getProductDataInvalidAtIndex( 1, EMPTY ),    // name
            productAndStockMasterDataInitializer.getProductDataInvalidAtIndex( 2, NULL ),     // description
            productAndStockMasterDataInitializer.getProductDataInvalidAtIndex( 2, EMPTY ),    // description
            productAndStockMasterDataInitializer.getProductDataInvalidAtIndex( 4, NULL ),     // buyingPrice
            productAndStockMasterDataInitializer.getProductDataInvalidAtIndex( 5, NULL ),     // sellingPrice
        };

        // when
        // then
        for ( Object[] invalidProductDataInstance : invalidProductData ) {
            StringBuffer invalidDataString = new StringBuffer().append( invalidProductDataInstance[0] )
                    .append( ", " ).append( invalidProductDataInstance[1] ).append( ", " ).append( invalidProductDataInstance[2] )
                    .append( ", " ).append( invalidProductDataInstance[3] ).append( ", " ).append( invalidProductDataInstance[4] )
                    .append( ", " ).append( invalidProductDataInstance[5] );
            assertThrows( ShopException.class, () -> productAndStockMasterDataInitializer.addProductDataToCatalog( invalidProductDataInstance ),
                    "Invalid data: " + invalidDataString.toString() );
        }
    }


    @Test
    public void testRemoveProductFromCatalog() {
        // given
        UUID productId = (UUID) PRODUCT_DATA[4][0];

        // when
        assertDoesNotThrow( () -> productCatalogUseCases.getSellingPrice( productId ) );
        productCatalogUseCases.removeProductFromCatalog( productId );

        // then
        assertThrows( ShopException.class, () ->
                productCatalogUseCases.getSellingPrice( productId ) );
    }



    @Test
    public void testRemoveNonExistentProduct() {
        // given
        UUID nonExistentProductId = UUID.randomUUID();

        // when
        // then
        assertThrows( ShopException.class,
                () -> productCatalogUseCases.removeProductFromCatalog( nonExistentProductId ) );
    }



    @Test
    public void testRemoveProductThatIsInStock() {
        // given
        UUID productId = (UUID) PRODUCT_DATA[4][0];
        UUID depotId0 = DEPOT_ID[0];
        depotUseCases.addToStock( depotId0, productId, 3 );

        // when
        // then
        assertThrows( ShopException.class,
                () -> productCatalogUseCases.removeProductFromCatalog( productId ) );
    }


    @Test
    public void testRemoveProductThatIsInShoppingCartOrOrder() {
        // given
        UUID productId3 = (UUID) PRODUCT_DATA[3][0];
        UUID productId4 = (UUID) PRODUCT_DATA[4][0];
        UUID depotId0 = DEPOT_ID[0];
        EmailAddressType userEmailAddress3 = USER_EMAIL[3];
        EmailAddressType userEmailAddress4 = USER_EMAIL[4];
        depotUseCases.addToStock( depotId0, productId3, 3 );
        depotUseCases.addToStock( depotId0, productId4, 4 );

        // when
        shoppingCartUseCases.addProductToShoppingCart(
                userEmailAddress3, productId3, 3 );
        shoppingCartUseCases.addProductToShoppingCart(
                userEmailAddress4, productId4, 4 );
        shoppingCartUseCases.checkout( userEmailAddress4 );

        // then
        assertThrows( ShopException.class,
                () -> productCatalogUseCases.removeProductFromCatalog( productId3 ) );
        assertThrows( ShopException.class,
                () -> productCatalogUseCases.removeProductFromCatalog( productId4 ) );
    }


    @Test
    public void testClearProductCatalog() {
        // given
        // when
        productCatalogUseCases.deleteProductCatalog();

        // then
        assertThrows( ShopException.class,
                () -> productCatalogUseCases.getSellingPrice( (UUID) PRODUCT_DATA[4][0] ) );
    }

}
