package thkoeln.archilab.ecommerce.test.regression;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import thkoeln.archilab.ecommerce.ShopException;
import thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer;
import thkoeln.archilab.ecommerce.usecases.masterdata.UserMasterDataInitializer;
import thkoeln.archilab.ecommerce.usecases.masterdata.Purgatory;
import thkoeln.archilab.ecommerce.usecases.*;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.EmailAddressType;

import java.util.List;
import java.util.Map;
import java.util.UUID;

import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer.PRODUCT_DATA;
import static thkoeln.archilab.ecommerce.usecases.masterdata.UserMasterDataInitializer.USER_EMAIL;

/**
 * This test checks the lifecycle of shipment units, so that e.g.
 * an order is not shipped twice.
 */
@SpringBootTest
public class ShipmentUnitsLifecycleTest {
    @Autowired
    private UserRegistrationUseCases userRegistrationUseCases;
    @Autowired
    private ShoppingCartUseCases shoppingCartUseCases;
    @Autowired
    private ProductCatalogUseCases productCatalogUseCases;
    @Autowired
    private DepotUseCases depotUseCases;
    @Autowired
    private ShipmentUnitUseCases shipmentUnitUseCases;
    @Autowired
    private Purgatory purgatory;

    private UserMasterDataInitializer userMasterDataInitializer;
    private ProductAndStockMasterDataInitializer productAndStockMasterDataInitializer;

    private EmailAddressType userEmailAddress;

    @BeforeEach
    public void setUp() {
        purgatory.deleteEverything();

        userMasterDataInitializer = new UserMasterDataInitializer( userRegistrationUseCases );
        userMasterDataInitializer.registerAllUsers();

        productAndStockMasterDataInitializer = new ProductAndStockMasterDataInitializer(
                productCatalogUseCases, depotUseCases );
        productAndStockMasterDataInitializer.addAllProducts();
        productAndStockMasterDataInitializer.addAllDepots();
        productAndStockMasterDataInitializer.addAllStock();

        // put products in the shopping cart ...
        userEmailAddress = USER_EMAIL[2];
        shoppingCartUseCases.addProductToShoppingCart(
                userEmailAddress, (UUID) PRODUCT_DATA[3][0], 2 );
        shoppingCartUseCases.addProductToShoppingCart(
                userEmailAddress, (UUID) PRODUCT_DATA[2][0], 1 );
    }


    @Test
    public void testValidOrderHasValidShipmentUnits() {
        // given
        UUID orderId = shoppingCartUseCases.checkout( userEmailAddress );
        assertNotNull( orderId );

        // when
        List<UUID> depots =
                shipmentUnitUseCases.getContributingDepotsForOrder( orderId );

        // then
        assertNotNull( depots );
        assertTrue( depots.size() > 0 );
        Map<UUID, Integer> shipmentUnit =
                shipmentUnitUseCases.getShipmentUnitForOrderAndDepot(
                        orderId, depots.get( 0 ) );
        assertNotNull( shipmentUnit );
        assertTrue( shipmentUnit.size() > 0 );
    }


    @Test
    public void testInvalidParameters() {
        // given
        // when
        assertThrows( ShopException.class,
                () -> shipmentUnitUseCases.getContributingDepotsForOrder( UUID.randomUUID() ) );
        assertThrows( ShopException.class,
                () -> shipmentUnitUseCases.getContributingDepotsForOrder( null ) );
        assertThrows( ShopException.class,
                () -> shipmentUnitUseCases.
                        getShipmentUnitForOrderAndDepot( null, null ) );
        assertThrows( ShopException.class,
                () -> shipmentUnitUseCases.
                        getShipmentUnitForOrderAndDepot( null, UUID.randomUUID() ) );
        assertThrows( ShopException.class,
                () -> shipmentUnitUseCases.
                        getShipmentUnitForOrderAndDepot( UUID.randomUUID(), null ) );
        assertThrows( ShopException.class,
                () -> shipmentUnitUseCases.
                        getShipmentUnitForOrderAndDepot( UUID.randomUUID(), UUID.randomUUID() ) );
    }
}
