package thkoeln.archilab.ecommerce.test.regression;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import thkoeln.archilab.ecommerce.ShopException;
import thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer;
import thkoeln.archilab.ecommerce.usecases.masterdata.UserMasterDataInitializer;
import thkoeln.archilab.ecommerce.usecases.masterdata.Purgatory;
import thkoeln.archilab.ecommerce.usecases.*;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.EmailAddressType;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.MoneyType;

import java.util.Map;
import java.util.UUID;

import static org.junit.jupiter.api.Assertions.*;
import static thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer.*;
import static thkoeln.archilab.ecommerce.usecases.masterdata.UserMasterDataInitializer.USER_EMAIL;
import static thkoeln.archilab.ecommerce.usecases.masterdata.FactoryMethodInvoker.instantiateEmailAddress;

@SpringBootTest
public class ShoppingCartTest {
    @Autowired
    private UserRegistrationUseCases userRegistrationUseCases;
    @Autowired
    private ShoppingCartUseCases shoppingCartUseCases;
    @Autowired
    private ProductCatalogUseCases productCatalogUseCases;
    @Autowired
    private DepotUseCases depotUseCases;
    @Autowired
    private OrderUseCases orderUseCases;
    @Autowired
    private ShipmentUnitUseCases shipmentUnitUseCases;
    @Autowired
    private Purgatory purgatory;

    private EmailAddressType nonExistingEmailAddress;

    private UserMasterDataInitializer userMasterDataInitializer;
    private ProductAndStockMasterDataInitializer productAndStockMasterDataInitializer;

    @BeforeEach
    public void setUp() {
        purgatory.deleteEverything();

        userMasterDataInitializer = new UserMasterDataInitializer( userRegistrationUseCases );
        userMasterDataInitializer.registerAllUsers();

        productAndStockMasterDataInitializer = new ProductAndStockMasterDataInitializer(
                productCatalogUseCases, depotUseCases );
        productAndStockMasterDataInitializer.addAllProducts();
        productAndStockMasterDataInitializer.addAllDepots();
        productAndStockMasterDataInitializer.addAllStock();

        nonExistingEmailAddress = instantiateEmailAddress( "this@nononono.de" );
    }


    @Test
    public void testInvalidRemoveFromShoppingCart() {
        // given
        UUID nonExistentProductId = UUID.randomUUID();
        UUID productId5 = (UUID) PRODUCT_DATA[5][0];
        UUID productId0 = (UUID) PRODUCT_DATA[0][0];
        UUID productId1 = (UUID) PRODUCT_DATA[1][0];
        UUID productId2 = (UUID) PRODUCT_DATA[2][0];
        EmailAddressType userEmailAddress0 = USER_EMAIL[0];
        shoppingCartUseCases.addProductToShoppingCart(
                userEmailAddress0, productId1, 5 );
        shoppingCartUseCases.addProductToShoppingCart(
                userEmailAddress0, productId2, 15 );

        // when
        shoppingCartUseCases.removeProductFromShoppingCart(
                userEmailAddress0, productId1, 2 );
        shoppingCartUseCases.removeProductFromShoppingCart(
                userEmailAddress0, productId2, 4 );
        shoppingCartUseCases.removeProductFromShoppingCart(
                userEmailAddress0, productId2, 7 );

        // then
        assertThrows( ShopException.class,
                () -> shoppingCartUseCases.removeProductFromShoppingCart(
                        userEmailAddress0, nonExistentProductId, 12 ) );
        assertThrows( ShopException.class,
                () -> shoppingCartUseCases.removeProductFromShoppingCart(
                        nonExistingEmailAddress, productId5, 12 ) );
        assertThrows( ShopException.class,
                () -> shoppingCartUseCases.removeProductFromShoppingCart(
                        userEmailAddress0, productId5, -1 ) );
        assertThrows( ShopException.class,
                () -> shoppingCartUseCases.removeProductFromShoppingCart(
                        userEmailAddress0, productId0, 1 ) );
        assertThrows( ShopException.class,
                () -> shoppingCartUseCases.removeProductFromShoppingCart(
                        userEmailAddress0, productId1, 4 ) );
        assertThrows( ShopException.class,
                () -> shoppingCartUseCases.removeProductFromShoppingCart(
                        userEmailAddress0, productId2, 5 ) );
    }



    @Test
    public void testCheckout() {
        // given
        UUID productId1 = (UUID) PRODUCT_DATA[1][0];
        UUID productId2 = (UUID) PRODUCT_DATA[2][0];
        UUID productId3 = (UUID) PRODUCT_DATA[3][0];
        int stock1before = PRODUCT_STOCK.get( productId1 )[0];
        int stock2before = PRODUCT_STOCK.get( productId2 )[0];
        int stock3before = PRODUCT_STOCK.get( productId3 )[0];
        EmailAddressType userEmailAddress3 = USER_EMAIL[3];
        Map<UUID, Integer> orderHistoryBefore = orderUseCases.getOrderHistory( userEmailAddress3 );

        // when
        shoppingCartUseCases.addProductToShoppingCart( userEmailAddress3, productId1,
                stock1before-2 );
        shoppingCartUseCases.addProductToShoppingCart( userEmailAddress3, productId2, 4 );
        shoppingCartUseCases.addProductToShoppingCart( userEmailAddress3, productId3, 5 );
        shoppingCartUseCases.checkout( userEmailAddress3 );
        int stock1after = depotUseCases.getAvailableStock( productId1 );
        int stock2after = depotUseCases.getAvailableStock( productId2 );
        int stock3after = depotUseCases.getAvailableStock( productId3 );
        Map<UUID, Integer> orderHistoryAfter = orderUseCases.getOrderHistory( userEmailAddress3 );

        // then
        assertEquals( 0, orderHistoryBefore.size() );
        assertEquals( 2, stock1after );
        assertEquals( stock2before-4, stock2after );
        assertEquals( stock3before-5, stock3after );
        assertEquals( 3, orderHistoryAfter.size() );
        assertEquals( stock1before-2, orderHistoryAfter.get( productId1 ) );
        assertEquals( 4, orderHistoryAfter.get( productId2 ) );
        assertEquals( 5, orderHistoryAfter.get( productId3 ) );
    }


    @Test
    public void testCheckoutInvalid() {
        // given
        EmailAddressType userEmailAddress3 = USER_EMAIL[3];
        EmailAddressType userEmailAddress5 = USER_EMAIL[5];
        UUID productId2 = (UUID) PRODUCT_DATA[2][0];

        // when
        shoppingCartUseCases.addProductToShoppingCart( userEmailAddress5, productId2, 4 );
        shoppingCartUseCases.removeProductFromShoppingCart( userEmailAddress5, productId2, 4 );

        // then
        assertThrows( ShopException.class, () -> shoppingCartUseCases.checkout( nonExistingEmailAddress ) );
        assertThrows( ShopException.class, () -> shoppingCartUseCases.checkout( userEmailAddress3 ) );
        assertThrows( ShopException.class, () -> shoppingCartUseCases.checkout( userEmailAddress5 ) );
    }


}
