package thkoeln.archilab.ecommerce.test.regression;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import thkoeln.archilab.ecommerce.ShopException;
import thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer;
import thkoeln.archilab.ecommerce.usecases.masterdata.UserMasterDataInitializer;
import thkoeln.archilab.ecommerce.usecases.masterdata.Purgatory;
import thkoeln.archilab.ecommerce.usecases.*;

import java.util.UUID;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer.*;

/**
 * This test checks the management of stock across depots.
 */
@SpringBootTest
public class StockTest {
    @Autowired
    private UserRegistrationUseCases userRegistrationUseCases;
    @Autowired
    private ProductCatalogUseCases productCatalogUseCases;
    @Autowired
    private DepotUseCases depotUseCases;
    @Autowired
    private Purgatory purgatory;

    private UserMasterDataInitializer userMasterDataInitializer;
    private ProductAndStockMasterDataInitializer productAndStockMasterDataInitializer;
    private UUID depotId0; // id of the first depot, where all products type 0 - 6 are stored
    private UUID depotId9; // the empty depot
    UUID nonExistingId1;
    UUID nonExistingId2;
    
    @BeforeEach
    public void setUp() {
        purgatory.deleteEverything();

        userMasterDataInitializer = new UserMasterDataInitializer( userRegistrationUseCases );
        userMasterDataInitializer.registerAllUsers();

        productAndStockMasterDataInitializer = new ProductAndStockMasterDataInitializer(
                productCatalogUseCases, depotUseCases );
        productAndStockMasterDataInitializer.addAllProducts();
        productAndStockMasterDataInitializer.addAllDepots();

        depotId0 = DEPOT_ID[0];
        depotId9 = DEPOT_ID[9];
        nonExistingId1 = UUID.randomUUID();
        nonExistingId2 = UUID.randomUUID();
    }


    @Test
    public void testAddToStock() {
        // given
        productAndStockMasterDataInitializer.addAllStock();
        UUID productId5 = (UUID) PRODUCT_DATA[5][0];

        // when
        int stock5before = depotUseCases.getAvailableStock( productId5 );
        assertEquals( PRODUCT_STOCK.get( productId5 )[0], stock5before );
        depotUseCases.addToStock( depotId0, productId5, 23 );
        int stock5after = depotUseCases.getAvailableStock( productId5 );
        depotUseCases.addToStock( depotId0, productId5, 0 );
        int stock5after2 = depotUseCases.getAvailableStock( productId5 );
        int stock5after3 = depotUseCases.getAvailableStock(
                depotId0, productId5 );

        // then
        assertEquals( stock5before + 23, stock5after );
        assertEquals( stock5after, stock5after2 );
        assertEquals( stock5after, stock5after3 );
    }


    @Test
    public void testInvalidAddToStock() {
        // given
        productAndStockMasterDataInitializer.addAllStock();
        UUID productId2 = (UUID) PRODUCT_DATA[2][0];

        // when
        // then
        assertThrows( ShopException.class,
                () -> depotUseCases.addToStock( nonExistingId1, nonExistingId2, 12 ) );
        assertThrows( ShopException.class,
                () -> depotUseCases.addToStock(
                        depotId0, nonExistingId2, 12 ) );
        assertThrows( ShopException.class,
                () -> depotUseCases.addToStock(
                        nonExistingId1, productId2, 12 ) );
        assertThrows( ShopException.class,
                () -> depotUseCases.addToStock(
                        depotId0, productId2, -1 ) );
    }


    @Test
    public void testRemoveFromStock() {
        // given
        productAndStockMasterDataInitializer.addAllStock();
        UUID productId5 = (UUID) PRODUCT_DATA[5][0];
        int stock5before = PRODUCT_STOCK.get( productId5 )[0];
        UUID productId6 = (UUID) PRODUCT_DATA[6][0];
        int stock6before = PRODUCT_STOCK.get( productId6 )[0];
        UUID productId0 = (UUID) PRODUCT_DATA[0][0];

        // when
        depotUseCases.removeFromStock( depotId0, productId5, 1 );
        int stock5after = depotUseCases.getAvailableStock( productId5 );
        depotUseCases.removeFromStock( depotId0, productId0, 0 );
        int stock0after = depotUseCases.getAvailableStock( productId0 );
        depotUseCases.removeFromStock(
                depotId0, productId6, stock6before );
        int stock6after = depotUseCases.getAvailableStock( productId6 );
        int stock6after2 = depotUseCases.getAvailableStock(
                depotId0, productId6 );
        // then
        assertEquals( stock5before - 1, stock5after );
        assertEquals( 0, stock0after );
        assertEquals( 0, stock6after );
        assertEquals( stock6after, stock6after2 );
    }


    @Test
    public void testInvalidRemoveFromStock() {
        // given
        productAndStockMasterDataInitializer.addAllStock();
        UUID productId5 = (UUID) PRODUCT_DATA[5][0];
        int stock5before = PRODUCT_STOCK.get( productId5 )[0];
        UUID productId0 = (UUID) PRODUCT_DATA[0][0];

        // when
        // then
        assertThrows( ShopException.class,
                () -> depotUseCases.removeFromStock( nonExistingId1, nonExistingId2, 12 ) );
        assertThrows( ShopException.class,
                () -> depotUseCases.removeFromStock(
                        depotId0, nonExistingId2, 12 ) );
        assertThrows( ShopException.class,
                () -> depotUseCases.removeFromStock(
                        nonExistingId1, productId5, 12 ) );
        assertThrows( ShopException.class,
                () -> depotUseCases.removeFromStock(
                        depotId0, productId5, -1 ) );
        assertThrows( ShopException.class,
                () -> depotUseCases.removeFromStock(
                        depotId0, productId5, stock5before + 1 ) );
        assertThrows( ShopException.class,
                () -> depotUseCases.removeFromStock(
                        depotId0, productId0, 1 ) );
    }


    @Test
    public void testChangeStock() {
        // given
        productAndStockMasterDataInitializer.addAllStock();
        UUID productId5 = (UUID) PRODUCT_DATA[5][0];

        // when
        depotUseCases.changeStockTo( depotId0, productId5, 111 );
        int stock10after = depotUseCases.getAvailableStock( productId5 );

        // then
        assertEquals( 111, stock10after );
    }


    @Test
    public void testInvalidChangeStock() {
        // given
        productAndStockMasterDataInitializer.addAllStock();
        UUID productId6 = (UUID) PRODUCT_DATA[6][0];

        // when
        // then
        assertThrows( ShopException.class, () -> depotUseCases.changeStockTo(
                nonExistingId1, nonExistingId2, 12 ) );
        assertThrows( ShopException.class, () -> depotUseCases.changeStockTo(
                nonExistingId1, productId6, 12 ) );
        assertThrows( ShopException.class, () -> depotUseCases.changeStockTo(
                depotId0, nonExistingId2, 12 ) );
        assertThrows( ShopException.class, () -> depotUseCases.changeStockTo(
                depotId0, productId6, -1 ) );
    }


    @Test
    public void testEmptyDepot() {
        // given
        productAndStockMasterDataInitializer.addAllStock();

        // when
        int totalStock = 0;
        for ( int i = 0; i < PRODUCT_STOCK.size(); i++ ) {
            UUID productId = (UUID) PRODUCT_DATA[i][0];
            totalStock += depotUseCases.getAvailableStock(
                    depotId9, productId );
        }

        // then
        assertEquals( 0, totalStock );
    }
}
