package thkoeln.archilab.ecommerce.test.shipmentunit;

import com.fasterxml.jackson.databind.ObjectMapper;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import thkoeln.archilab.ecommerce.test.ShoppingCartRESTHelper;
import thkoeln.archilab.ecommerce.test.restdtos.DepotIdDTO;
import thkoeln.archilab.ecommerce.usecases.ProductCatalogUseCases;
import thkoeln.archilab.ecommerce.usecases.UserRegistrationUseCases;
import thkoeln.archilab.ecommerce.usecases.DepotUseCases;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.EmailAddressType;
import thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer;
import thkoeln.archilab.ecommerce.usecases.masterdata.UserMasterDataInitializer;
import thkoeln.archilab.ecommerce.usecases.masterdata.Purgatory;

import javax.transaction.Transactional;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.jsonPath;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;
import static thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer.PRODUCT_DATA;
import static thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer.DEPOT_ID;
import static thkoeln.archilab.ecommerce.usecases.masterdata.UserMasterDataInitializer.USER_EMAIL;


@SpringBootTest
@Transactional
@AutoConfigureMockMvc
@ActiveProfiles("test")
public class ShipmentUnitRESTTest {
    @Autowired
    private UserRegistrationUseCases userRegistrationUseCases;
    @Autowired
    private ProductCatalogUseCases productCatalogUseCases;
    @Autowired
    private DepotUseCases depotUseCases;
    @Autowired
    private Purgatory purgatory;
    @Autowired
    private MockMvc mockMvc;

    private ShoppingCartRESTHelper shoppingCartRESTHelper;
    private Map<UUID, Integer>  map8_11_14_quantity_2_2_2, map8_11_14_quantity_3_3_4,
            map10_12_quantity_1_1, map11_quantity_1, map12_quantity_10,
            map8_9_10_11_quantity_2_1_4_2;

    private UserMasterDataInitializer userMasterDataInitializer;
    private ProductAndStockMasterDataInitializer productAndStockMasterDataInitializer;


    @BeforeEach
    public void setUp() {
        shoppingCartRESTHelper = new ShoppingCartRESTHelper(
                mockMvc, productCatalogUseCases, depotUseCases );
        purgatory.deleteEverything();

        userMasterDataInitializer = new UserMasterDataInitializer( userRegistrationUseCases );
        userMasterDataInitializer.registerAllUsers();

        productAndStockMasterDataInitializer = new ProductAndStockMasterDataInitializer(
                productCatalogUseCases, depotUseCases );
        productAndStockMasterDataInitializer.addAllProducts();
        productAndStockMasterDataInitializer.addAllDepots();
        productAndStockMasterDataInitializer.addAllStock();

        map8_11_14_quantity_2_2_2 = new HashMap<>() {{
            put( (UUID) PRODUCT_DATA[8][0], 2 );
            put( (UUID) PRODUCT_DATA[11][0], 2 );
            put( (UUID) PRODUCT_DATA[14][0], 2 );
        }};
        map8_11_14_quantity_3_3_4 = new HashMap<>() {{
            put( (UUID) PRODUCT_DATA[8][0], 3 );
            put( (UUID) PRODUCT_DATA[11][0], 3 );
            put( (UUID) PRODUCT_DATA[14][0], 4 );
        }};
        map10_12_quantity_1_1 = new HashMap<>() {{
            put( (UUID) PRODUCT_DATA[10][0], 1 );
            put( (UUID) PRODUCT_DATA[12][0], 1 );
        }};
        map11_quantity_1 = new HashMap<>() {{
            put( (UUID) PRODUCT_DATA[11][0], 1 );
        }};
        map12_quantity_10 = new HashMap<>() {{
            put( (UUID) PRODUCT_DATA[12][0], 10 );
        }};
        map8_9_10_11_quantity_2_1_4_2 = new HashMap<>() {{
            put( (UUID) PRODUCT_DATA[8][0], 2 );
            put( (UUID) PRODUCT_DATA[9][0], 1 );
            put( (UUID) PRODUCT_DATA[10][0], 4 );
            put( (UUID) PRODUCT_DATA[11][0], 2 );
        }};
    }


    @Test
    public void testInvalidUris() throws Exception {
        // given
        String allshipmentUnitsUri = "/shipmentUnits";
        String randomIdUri = "/shipmentUnits/" + UUID.randomUUID();
        String randomOrderUri = "/shipmentUnits?orderId=" + UUID.randomUUID();
        String randomParamUri = "/shipmentUnits?randomParam=randomValue";

        // when
        // then
        mockMvc.perform( get( allshipmentUnitsUri ) ).andExpect( status().isMethodNotAllowed() );
        mockMvc.perform( get( randomIdUri ) ).andExpect( status().isNotFound() );
        mockMvc.perform( get( randomOrderUri ) ).andExpect( status().isNotFound() );
        mockMvc.perform( get( randomParamUri ) ).andExpect( status().is4xxClientError() );
    }


    /**
     * Different users always buy the same product (no. 7). The product is available in
     * depot no. {1, 2, 3}. The test checks if the correct (closest) depot is
     * selected for each user. For user 0 and 3, there is only one solution. These two
     * are tested here.
     */
    @Test
    public void testClosestShipmentUnit() throws Exception {
        // given
        UUID productId7 = (UUID) PRODUCT_DATA[7][0];
        EmailAddressType userEmailAddress0 = USER_EMAIL[0];
        UUID shoppingCartId0 = shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress0 );
        EmailAddressType userEmailAddress3 = USER_EMAIL[3];
        UUID shoppingCartId3 = shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress3 );
        Map<UUID, Map<UUID, Integer>> depotMap0 = Map.of(
                DEPOT_ID[2], Map.of( productId7, 1 )
        );
        Map<UUID, Map<UUID, Integer>> depotMap3 = Map.of(
                DEPOT_ID[3], Map.of( productId7, 1 )
        );

        // when
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId0, productId7, 1 );
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId3, productId7, 1 );
        UUID orderId0 = shoppingCartRESTHelper.checkout( shoppingCartId0 );
        UUID orderId3 = shoppingCartRESTHelper.checkout( shoppingCartId3 );

        // then
        checkShipmentUnit( orderId0, depotMap0 );
        checkShipmentUnit( orderId3, depotMap3 );
    }




    @Test
    public void testClosestSingleShipmentUnitsWins() throws Exception {
        // given
        EmailAddressType userEmail3 = USER_EMAIL[3];
        UUID shoppingCartId3 = shoppingCartRESTHelper.getQueryShoppingCart( userEmail3 );
        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId3, (UUID) PRODUCT_DATA[8][0], 2 );
        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId3, (UUID) PRODUCT_DATA[11][0], 2 );
        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId3, (UUID) PRODUCT_DATA[14][0], 2 );
        // that basket could have been served from depot 4 or 7, but 4 is closer to the user
        UUID orderId = shoppingCartRESTHelper.checkout( shoppingCartId3 );

        // when
        Map<UUID, Map<UUID, Integer>> depotMap = Map.of(
                DEPOT_ID[4], map8_11_14_quantity_2_2_2
        );

        // then
        checkShipmentUnit( orderId, depotMap );
    }


    /**
     * Hint: (see comment on the first test)
     */
    @Test
    public void testDepotWithEnoughCapacityWins() throws Exception {
        // given
        EmailAddressType userEmail3 = USER_EMAIL[3];
        UUID shoppingCartId3 = shoppingCartRESTHelper.getQueryShoppingCart( userEmail3 );

        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId3, (UUID) PRODUCT_DATA[8][0], 3 );
        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId3, (UUID) PRODUCT_DATA[11][0], 3 );
        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId3, (UUID) PRODUCT_DATA[14][0], 4 );
        // that basket could have been served from depot 4 or 7, 4 is closer to the user,
        // but only 7 has enough capacity
        UUID orderId = shoppingCartRESTHelper.checkout( shoppingCartId3 );

        // when
        Map<UUID, Map<UUID, Integer>> depotMap = Map.of(
                DEPOT_ID[7], map8_11_14_quantity_3_3_4
        );

        // then
        checkShipmentUnit( orderId, depotMap );
    }


    /**
     * Hint: (see comment on the first test)
     */
    @Test
    public void testTwoShipmentUnits() throws Exception {
        // given
        EmailAddressType userEmail6 = USER_EMAIL[6];
        UUID shoppingCartId6 = shoppingCartRESTHelper.getQueryShoppingCart( userEmail6 );

        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId6, (UUID) PRODUCT_DATA[10][0], 1 );
        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId6, (UUID) PRODUCT_DATA[11][0], 1 );
        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId6, (UUID) PRODUCT_DATA[12][0], 1 );
        // that basket needs two shipment units (2 + 1 products) anyway, and there
        // are 2 options for the bigger one: 10+12 from 5, or 10+11 from 7. The first one is
        // closer. The smaller can then be served from 4 (closest), 7, or 8.
        UUID orderId = shoppingCartRESTHelper.checkout( shoppingCartId6 );

        // when
        Map<UUID, Map<UUID, Integer>> depotMap = Map.of(
                DEPOT_ID[5], map10_12_quantity_1_1,
                DEPOT_ID[4], map11_quantity_1
        );

        // then
        checkShipmentUnit( orderId, depotMap );
    }


    /**
     * Hint: (see comment on the first test)
     */
    @Test
    public void testTwoBigShipmentUnits() throws Exception {
        // given
        EmailAddressType userEmail2 = USER_EMAIL[2];
        UUID shoppingCartId2 = shoppingCartRESTHelper.getQueryShoppingCart( userEmail2 );

        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId2, (UUID) PRODUCT_DATA[8][0], 2 );
        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId2, (UUID) PRODUCT_DATA[9][0], 1 );
        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId2, (UUID) PRODUCT_DATA[10][0], 4 );
        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId2, (UUID) PRODUCT_DATA[11][0], 2 );
        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId2, (UUID) PRODUCT_DATA[12][0], 10 );
        UUID orderId = shoppingCartRESTHelper.checkout( shoppingCartId2 );

        // when
        Map<UUID, Map<UUID, Integer>> depotMap = Map.of(
                DEPOT_ID[7], map8_9_10_11_quantity_2_1_4_2,
                DEPOT_ID[5], map12_quantity_10
        );

        // then
        checkShipmentUnit( orderId, depotMap );
    }


    /**
     * Hint: (see comment on the first test)
     */
    @Test
    public void testOnlyOneSolution() throws Exception {
        // given
        EmailAddressType userEmail6 = USER_EMAIL[6];
        UUID shoppingCartId3 = shoppingCartRESTHelper.getQueryShoppingCart( userEmail6 );

        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId3, (UUID) PRODUCT_DATA[12][0], 10 );
        UUID orderId = shoppingCartRESTHelper.checkout( shoppingCartId3 );

        // when
        Map<UUID, Map<UUID, Integer>> depotMap = Map.of(
                DEPOT_ID[5], map12_quantity_10
        );

        // then
        checkShipmentUnit( orderId, depotMap );
    }


    /**
     * Calls the /shipmentUnits endpoint and checks the response.
     * @param orderId The id of the order.
     * @param depotMap The expected depots and their quantities, as a map of maps:
     *                            depots -> ( products -> quantity )
     * @throws Exception
     */
    public void checkShipmentUnit( UUID orderId,
                                                  Map<UUID, Map<UUID, Integer>> depotMap ) throws Exception {
        // call the GET endpoint
        String shipmentUnitUri = "/shipmentUnits?orderId=" + orderId.toString();
        ResultActions resultActions = mockMvc.perform( get( shipmentUnitUri ) )
                .andExpect( status().isOk() );

        // check the response
        ObjectMapper objectMapper = new ObjectMapper();
        resultActions.andExpect( jsonPath( "$.length()" ).value( depotMap.size() ) );
        for ( int i = 0; i < depotMap.size(); i++ ) {
            resultActions.andExpect( jsonPath( "$[" + i + "].id" ).exists() );
        }
        DepotIdDTO[] depotIdDTOs = objectMapper.readValue(
                resultActions.andReturn().getResponse().getContentAsString(), DepotIdDTO[].class );
        int i = 0;
        for ( DepotIdDTO depotIdDTO : depotIdDTOs ) {
            UUID depotId = depotIdDTO.getDepotId();
            Map<UUID, Integer> quantityMap = depotMap.get( depotId );
            int numOfProducts = quantityMap.size();
            String jsonPathLengthExpr = "$[" + i + "].shipmentUnitParts.length()";
            resultActions.andExpect( jsonPath( jsonPathLengthExpr ).value( numOfProducts ) );
            for ( Map.Entry<UUID, Integer> entry : quantityMap.entrySet() ) {
                UUID productId = entry.getKey();
                Integer quantity = entry.getValue();
                resultActions.andExpect( jsonPath( "$[" + i + "].shipmentUnitParts[?(@.productId == '"
                        + productId + "')].quantity" ).value( quantity ) );
            }
            i++;
        }
    }
}
