package thkoeln.archilab.ecommerce.test.shoppingcart;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultMatcher;
import thkoeln.archilab.ecommerce.test.ShoppingCartRESTHelper;
import thkoeln.archilab.ecommerce.usecases.ShoppingCartUseCases;
import thkoeln.archilab.ecommerce.usecases.ProductCatalogUseCases;
import thkoeln.archilab.ecommerce.usecases.UserRegistrationUseCases;
import thkoeln.archilab.ecommerce.usecases.DepotUseCases;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.EmailAddressType;
import thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer;
import thkoeln.archilab.ecommerce.usecases.masterdata.UserMasterDataInitializer;
import thkoeln.archilab.ecommerce.usecases.masterdata.Purgatory;

import javax.transaction.Transactional;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

import static org.junit.jupiter.api.Assertions.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;
import static thkoeln.archilab.ecommerce.usecases.masterdata.ProductAndStockMasterDataInitializer.*;
import static thkoeln.archilab.ecommerce.usecases.masterdata.UserMasterDataInitializer.USER_EMAIL;
import static thkoeln.archilab.ecommerce.usecases.masterdata.FactoryMethodInvoker.instantiateEmailAddress;


@SpringBootTest
@Transactional
@AutoConfigureMockMvc
@ActiveProfiles("test")
public class ShoppingCartRESTTest {
    @Autowired
    private UserRegistrationUseCases userRegistrationUseCases;
    @Autowired
    private ProductCatalogUseCases productCatalogUseCases;
    @Autowired
    private DepotUseCases depotUseCases;
    @Autowired
    private ShoppingCartUseCases shoppingCartUseCases;
    @Autowired
    private Purgatory purgatory;
    @Autowired
    private MockMvc mockMvc;

    private ShoppingCartRESTHelper shoppingCartRESTHelper;
    private UserMasterDataInitializer userMasterDataInitializer;
    private ProductAndStockMasterDataInitializer productAndStockMasterDataInitializer;
    private EmailAddressType nonExistingEmailAddress;

    private static final ResultMatcher NOT_FOUND = status().isNotFound();
    private static final ResultMatcher OK = status().isOk();
    private static final ResultMatcher CREATED = status().isCreated();
    private static final ResultMatcher UNPROCESSABLE_ENTITY = status().isUnprocessableEntity();
    private static final ResultMatcher CONFLICT = status().isConflict();

    @BeforeEach
    public void setUp() {
        shoppingCartRESTHelper = new ShoppingCartRESTHelper(
                mockMvc, productCatalogUseCases, depotUseCases );
        purgatory.deleteEverything();

        userMasterDataInitializer = new UserMasterDataInitializer( userRegistrationUseCases );
        userMasterDataInitializer.registerAllUsers();

        productAndStockMasterDataInitializer = new ProductAndStockMasterDataInitializer(
                productCatalogUseCases, depotUseCases );
        productAndStockMasterDataInitializer.addAllProducts();
        productAndStockMasterDataInitializer.addAllDepots();
        productAndStockMasterDataInitializer.addAllStock();

        nonExistingEmailAddress = instantiateEmailAddress( "harry@sally.de" );
    }


    @Test
    public void testInvalidShoppingCartBaseUris() throws Exception {
        // given
        String allShoppingCartsUri = "/shoppingCarts";
        String randomIdUri = "/shoppingCarts/" + UUID.randomUUID();
        String randomUserUUID = "/shoppingCarts?userId=" + UUID.randomUUID();
        String invalidUserUUID = "/shoppingCarts?userId=invalidUUID";
        String randomParamUri = "/shoppingCarts?randomParam=randomValue";

        // when
        // then
        mockMvc.perform( get( allShoppingCartsUri ) ).andExpect( status().isMethodNotAllowed() );
        mockMvc.perform( get( randomIdUri ) ).andExpect( status().isNotFound() );
        mockMvc.perform( get( randomUserUUID ) ).andExpect( status().isNotFound() );
        mockMvc.perform( get( invalidUserUUID ) ).andExpect( status().isNotFound() );
        mockMvc.perform( get( randomParamUri ) ).andExpect( status().is4xxClientError() );
    }


    @Test
    public void testInvalidShoppingCartPositionUris() throws Exception {
        // given
        EmailAddressType userEmailAddress8 = USER_EMAIL[8];
        UUID shoppingCartId8 = shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress8 );
        String randomShoppingCartPositionUri = "/shoppingCarts/" + shoppingCartId8.toString() + "/random";
        String randomProductUri = "/shoppingCarts/" + shoppingCartId8.toString()
                + "/positions/" + UUID.randomUUID();

        // when
        // then
        mockMvc.perform( post( randomShoppingCartPositionUri ) ).andExpect( status().isNotFound() );
        mockMvc.perform( delete( randomProductUri ) ).andExpect( status().isNotFound() );
    }


    @Test
    public void testInvalidAddToShoppingCart() throws Exception {
        // given
        UUID nonExistentProductId = UUID.randomUUID();
        UUID productId5 = (UUID) PRODUCT_DATA[5][0];
        UUID productId0 = (UUID) PRODUCT_DATA[0][0];
        UUID productId1 = (UUID) PRODUCT_DATA[1][0];
        UUID productId2 = (UUID) PRODUCT_DATA[2][0];
        EmailAddressType userEmailAddress0 = USER_EMAIL[0];
        Map<UUID, Integer> startQuantityMap = new HashMap<>();
        startQuantityMap.put( productId2, 19 );

        // when
        UUID shoppingCartId = shoppingCartRESTHelper.getQueryShoppingCart(
                userEmailAddress0, new HashMap<>() );
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId, productId2, 6 );
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId, productId2, 13 );

        // then
        shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress0, startQuantityMap );

        // ... and check a couple of invalid cases
        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId, nonExistentProductId, 12, NOT_FOUND );
        shoppingCartRESTHelper.addProductToShoppingCart(
                UUID.randomUUID(), productId5, 12, NOT_FOUND );
        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId, productId5, -1, UNPROCESSABLE_ENTITY );

        // productId0 is not available in the requested quantity (or rather, not at all)
        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId, productId0, 1, CONFLICT );

        // There are only 10 pieces of productId1 available
        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId, productId1, 11, CONFLICT );

        // There are only 20 pieces of productId1 available - we have already 19 in the shopping cart
        shoppingCartRESTHelper.addProductToShoppingCart(
                shoppingCartId, productId2, 2, CONFLICT );

        // check if the shopping cart is still the same
        shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress0, startQuantityMap );
    }


    @Test
    public void testDeleteProductFromShoppingCart() throws Exception {
        // given
        UUID productId1 = (UUID) PRODUCT_DATA[1][0];
        UUID productId2 = (UUID) PRODUCT_DATA[2][0];
        UUID productId3 = (UUID) PRODUCT_DATA[3][0];
        EmailAddressType userEmailAddress6 = USER_EMAIL[6];
        UUID shoppingCartId6 = shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress6 );
        Map<UUID, Integer> quantityMap1 = new HashMap<>();
        quantityMap1.put( productId1, 1 );
        quantityMap1.put( productId2, 2 );
        quantityMap1.put( productId3, 3 );
        Map<UUID, Integer> quantityMap2 = new HashMap<>();
        quantityMap2.put( productId1, 1 );
        quantityMap2.put( productId3, 3 );

        // when
        // then
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId6, productId1, 1 );
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId6, productId2, 2 );
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId6, productId3, 3 );
        shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress6, quantityMap1 );

        shoppingCartRESTHelper.deleteProductFromShoppingCart( shoppingCartId6, productId2 );
        shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress6, quantityMap2 );
    }


    @Test
    public void testNoDoubleDelete() throws Exception {
        // given
        UUID productId3 = (UUID) PRODUCT_DATA[3][0];
        EmailAddressType userEmailAddress6 = USER_EMAIL[6];
        UUID shoppingCartId6 = shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress6 );

        // when
        // then
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId6, productId3, 3 );
        shoppingCartRESTHelper.deleteProductFromShoppingCart( shoppingCartId6, productId3 );
        shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress6, new HashMap<>() );
        shoppingCartRESTHelper.deleteProductFromShoppingCart( shoppingCartId6, productId3, CONFLICT );
    }


    @Test
    public void testAddRemoveProductsFromAndToShoppingCart() throws Exception {
        // given
        UUID productId1 = (UUID) PRODUCT_DATA[1][0];
        UUID productId2 = (UUID) PRODUCT_DATA[2][0];
        EmailAddressType userEmailAddress3 = USER_EMAIL[3];
        EmailAddressType userEmailAddress5 = USER_EMAIL[5];
        UUID shoppingCartId3 = shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress3 );
        UUID shoppingCartId5 = shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress5 );

        // when
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId3, productId1, 2 );
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId3, productId2, 3 );
        shoppingCartRESTHelper.deleteProductFromShoppingCart( shoppingCartId3, productId1 );
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId3, productId1, 1 );
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId3, productId2, 6 );

        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId5, productId1, 2 );
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId5, productId2, 8 );
        shoppingCartRESTHelper.deleteProductFromShoppingCart( shoppingCartId5, productId1 );
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId5, productId1, 1 );
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId5, productId2, 2 );

        int reservedStock1 = shoppingCartUseCases.getReservedStockInShoppingCarts( productId1 );
        int reservedStock2 = shoppingCartUseCases.getReservedStockInShoppingCarts( productId2 );

        // then
        // user3 has 1x productId1 and 9x productId2 in cart
        Map<UUID, Integer> cart3 = new HashMap<>();
        cart3.put( productId1, 1 );
        cart3.put( productId2, 9 );
        shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress3, cart3 );

        // user5 has 1x productId1 and 10x productId2 in cart
        Map<UUID, Integer> cart5 = new HashMap<>();
        cart5.put( productId1, 1 );
        cart5.put( productId2, 10 );
        shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress5, cart5 );

        assertEquals( 2, reservedStock1 );
        assertEquals( 19, reservedStock2 );
    }


    @Test
    public void testImpactOfStockCorrectionToOneShoppingCart() throws Exception {
        // given
        UUID productId1 = (UUID) PRODUCT_DATA[1][0];
        UUID productId2 = (UUID) PRODUCT_DATA[2][0];
        UUID productId3 = (UUID) PRODUCT_DATA[3][0];
        UUID depotId0 = DEPOT_ID[0];
        EmailAddressType userEmailAddress3 = USER_EMAIL[3];
        UUID shoppingCartId3 = shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress3 );

        // when
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId3, productId1, 6 );
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId3, productId2, 15 );
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId3, productId3, 1 );
        depotUseCases.changeStockTo( depotId0, productId1, 4 );
        depotUseCases.changeStockTo( depotId0, productId2, 16 );
        depotUseCases.changeStockTo( depotId0, productId3, 0 );
        int reservedStock1 = shoppingCartUseCases.getReservedStockInShoppingCarts( productId1 );
        int reservedStock2 = shoppingCartUseCases.getReservedStockInShoppingCarts( productId2 );
        int reservedStock3 = shoppingCartUseCases.getReservedStockInShoppingCarts( productId3 );

        // then
        Map<UUID, Integer> cart3 = new HashMap<>();
        cart3.put( productId1, 4 );
        cart3.put( productId2, 15 );
        shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress3, cart3 );

        assertEquals( 4, reservedStock1 );
        assertEquals( 15, reservedStock2 );
        assertEquals( 0, reservedStock3 );
    }


    @Test
    public void testImpactOfStockCorrectionToSeveralShoppingCarts() throws Exception {
        // given
        UUID productId2 = (UUID) PRODUCT_DATA[2][0];
        EmailAddressType userEmailAddress3 = USER_EMAIL[3];
        EmailAddressType userEmailAddress6 = USER_EMAIL[6];
        EmailAddressType userEmailAddress9 = USER_EMAIL[9];
        UUID shoppingCartId3 = shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress3 );
        UUID shoppingCartId6 = shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress6 );
        UUID shoppingCartId9 = shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress9 );
        UUID depotId0 = DEPOT_ID[0];

        // when
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId3, productId2, 3 );
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId6, productId2, 6 );
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId9, productId2, 9 );
        depotUseCases.removeFromStock( depotId0, productId2, 2 );

        int reservedStock21 = shoppingCartUseCases.getReservedStockInShoppingCarts( productId2 );
        Map<UUID, Integer> cart3 = new HashMap<>();
        cart3.put( productId2, 3 );
        shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress3, cart3 );
        Map<UUID, Integer> cart6 = new HashMap<>();
        cart6.put( productId2, 6 );
        shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress6, cart6 );
        Map<UUID, Integer> cart9 = new HashMap<>();
        cart9.put( productId2, 9 );
        shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress9, cart9 );

        // for simplicity reasons, we use the use case interface here
        depotUseCases.removeFromStock( depotId0, productId2, 8 );
        Map<UUID, Integer> cart32 = shoppingCartUseCases.getShoppingCartAsMap( userEmailAddress3 );
        Map<UUID, Integer> cart62 = shoppingCartUseCases.getShoppingCartAsMap( userEmailAddress6 );
        Map<UUID, Integer> cart92 = shoppingCartUseCases.getShoppingCartAsMap( userEmailAddress9 );
        int quantity32 = cart32.get( productId2 ) == null ? 0 : cart32.get( productId2 );
        int quantity62 = cart62.get( productId2 ) == null ? 0 : cart62.get( productId2 );
        int quantity92 = cart92.get( productId2 ) == null ? 0 : cart92.get( productId2 );
        int reservedStock22 = shoppingCartUseCases.getReservedStockInShoppingCarts( productId2 );

        // then
        assertEquals( 18, reservedStock21 );
        assertEquals( 10, reservedStock22 );
        assertEquals( reservedStock22, quantity32 + quantity62 + quantity92 );
    }


    @Test
    public void testCheckoutCreatesOrder() throws Exception {
        // given
        UUID productId1 = (UUID) PRODUCT_DATA[1][0];
        UUID productId2 = (UUID) PRODUCT_DATA[2][0];
        EmailAddressType userEmailAddress9 = USER_EMAIL[9];
        UUID shoppingCartId9 = shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress9 );
        int stock1before = PRODUCT_STOCK.get( productId1 )[0];
        int stock2before = PRODUCT_STOCK.get( productId2 )[0];

        // when
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId9, productId1, 2 );
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId9, productId2, 3 );
        UUID orderId = shoppingCartRESTHelper.checkout( shoppingCartId9 );
        int stock1after = depotUseCases.getAvailableStock( productId1 );
        int stock2after = depotUseCases.getAvailableStock( productId2 );

        // then
        assertNotNull( orderId );
        assertEquals( 2, stock1before - stock1after );
        assertEquals( 3, stock2before - stock2after );
    }


    @Test
    public void testNoDoubleCheckout() throws Exception {
        // given
        EmailAddressType userEmailAddress9 = USER_EMAIL[9];
        UUID shoppingCartId9 = shoppingCartRESTHelper.getQueryShoppingCart( userEmailAddress9 );
        UUID productId2 = (UUID) PRODUCT_DATA[2][0];

        // when
        shoppingCartRESTHelper.addProductToShoppingCart( shoppingCartId9, productId2, 3 );

        // then
        shoppingCartRESTHelper.checkout( shoppingCartId9, CREATED );
        shoppingCartRESTHelper.checkout( shoppingCartId9, CONFLICT );
    }

}
