package thkoeln.archilab.ecommerce.test.user;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.web.servlet.MockMvc;
import thkoeln.archilab.ecommerce.usecases.UserRegistrationUseCases;
import thkoeln.archilab.ecommerce.usecases.UserType;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.EmailAddressType;
import thkoeln.archilab.ecommerce.usecases.domainprimitivetypes.PersonalAddressType;
import thkoeln.archilab.ecommerce.usecases.masterdata.UserMasterDataInitializer;
import thkoeln.archilab.ecommerce.usecases.masterdata.FactoryMethodInvoker;
import thkoeln.archilab.ecommerce.usecases.masterdata.Purgatory;

import javax.transaction.Transactional;
import java.util.UUID;

import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.put;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.jsonPath;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;
import static thkoeln.archilab.ecommerce.usecases.masterdata.UserMasterDataInitializer.*;

@SpringBootTest
@Transactional
@AutoConfigureMockMvc
@ActiveProfiles("test")
public class UserRESTTest {
    @Autowired
    private UserRegistrationUseCases userRegistrationUseCases;
    @Autowired
    private Purgatory purgatory;

    @Autowired
    private MockMvc mockMvc;

    private UserMasterDataInitializer userMasterDataInitializer;
    private EmailAddressType nonExistingEmailAddress;
    private UUID randomId;

    @BeforeEach
    public void setUp() {
        purgatory.deleteEverything();

        userMasterDataInitializer = new UserMasterDataInitializer( userRegistrationUseCases );
        userMasterDataInitializer.registerAllUsers();

        nonExistingEmailAddress = FactoryMethodInvoker.instantiateEmailAddress(
                "harry@sally.de" );
        randomId = UUID.randomUUID();
    }


    @Test
    public void testInvalidUserUris() throws Exception {
        // given
        String allUsersUri = "/users";
        String randomIdUri = "/users/" +  randomId;
        String nonExistingEmailAddressUri = "/users?emailAddress=" + nonExistingEmailAddress;
        String randomParamUri = "/users?randomParam=randomValue";

        // when
        // then
        mockMvc.perform( get( allUsersUri ) ).andExpect( status().isMethodNotAllowed() );
        mockMvc.perform( get( randomIdUri ) ).andExpect( status().isNotFound() );
        mockMvc.perform( get( nonExistingEmailAddressUri ) ).andExpect( status().isNotFound() );
        mockMvc.perform( get( randomParamUri ) ).andExpect( status().is4xxClientError() );
    }


    @Test
    public void testRandomUserId() throws Exception {
        // given
        String invalidUri = "/users/" +  randomId;

        // when
        // then
        mockMvc.perform( get( invalidUri ) ).andExpect( status().isNotFound() );
    }


    @Test
    public void testQueryUser() throws Exception {
        // given
        EmailAddressType emailAddress4 = USER_EMAIL[4];
        String userUri4 = "/users?emailAddress=" + emailAddress4;

        // when
        UserType user4 = userRegistrationUseCases.getUserData( emailAddress4 );
        String name = USER_NAME[4];
        PersonalAddressType personalAddress4 = USER_ADDRESS[4];

        // then (check using JSONPath)
        mockMvc.perform( get( userUri4 ) )
                .andExpect( status().isOk() )
                .andExpect( jsonPath( "$.name" ).value( name ) )
                // id is position 0 (cannot be tested with current API structure), name is position 1,
                // email address is 2, personal address is 3)
                .andExpect( jsonPath( "$..emailAddressString" ).value( emailAddress4.toString() ) )
                .andExpect( jsonPath( "$..city" ).value( personalAddress4.getCity() ) )
                .andExpect( jsonPath( "$..postalCodeString" ).value(
                        personalAddress4.getPostalCode().toString() ) );
    }

}
